#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from zepben.evolve.services.common.translator.base_cim2proto import identifiedobject_to_pb, organisationrole_to_pb, document_to_pb
from zepben.evolve.model import *
from zepben.evolve.services.common.translator.util import mrid_or_empty
from zepben.protobuf.cim.iec61968.assetinfo.CableInfo_pb2 import CableInfo as PBCableInfo
from zepben.protobuf.cim.iec61968.assetinfo.OverheadWireInfo_pb2 import OverheadWireInfo as PBOverheadWireInfo
from zepben.protobuf.cim.iec61968.assetinfo.WireInfo_pb2 import WireInfo as PBWireInfo
from zepben.protobuf.cim.iec61968.assetinfo.WireMaterialKind_pb2 import WireMaterialKind as PBWireMaterialKind
from zepben.protobuf.cim.iec61968.assets.AssetContainer_pb2 import AssetContainer as PBAssetContainer
from zepben.protobuf.cim.iec61968.assets.AssetInfo_pb2 import AssetInfo as PBAssetInfo
from zepben.protobuf.cim.iec61968.assets.AssetOrganisationRole_pb2 import AssetOrganisationRole as PBAssetOrganisationRole
from zepben.protobuf.cim.iec61968.assets.AssetOwner_pb2 import AssetOwner as PBAssetOwner
from zepben.protobuf.cim.iec61968.assets.Asset_pb2 import Asset as PBAsset
from zepben.protobuf.cim.iec61968.assets.Pole_pb2 import Pole as  PBPole
from zepben.protobuf.cim.iec61968.assets.StreetlightLampKind_pb2 import StreetlightLampKind as PBStreetlightLampKind
from zepben.protobuf.cim.iec61968.assets.Streetlight_pb2 import Streetlight as PBStreetlight
from zepben.protobuf.cim.iec61968.assets.Structure_pb2 import Structure as PBStructure
from zepben.protobuf.cim.iec61968.common.Location_pb2 import Location as PBLocation
from zepben.protobuf.cim.iec61968.common.PositionPoint_pb2 import PositionPoint as PBPositionPoint
from zepben.protobuf.cim.iec61968.common.StreetAddress_pb2 import StreetAddress as PBStreetAddress
from zepben.protobuf.cim.iec61968.common.TownDetail_pb2 import TownDetail as PBTownDetail
from zepben.protobuf.cim.iec61968.metering.EndDevice_pb2 import EndDevice as PBEndDevice
from zepben.protobuf.cim.iec61968.metering.Meter_pb2 import Meter as PBMeter
from zepben.protobuf.cim.iec61968.metering.UsagePoint_pb2 import UsagePoint as PBUsagePoint
from zepben.protobuf.cim.iec61968.operations.OperationalRestriction_pb2 import OperationalRestriction as PBOperationalRestriction
from zepben.protobuf.cim.iec61970.base.auxiliaryequipment.AuxiliaryEquipment_pb2 import AuxiliaryEquipment as PBAuxiliaryEquipment
from zepben.protobuf.cim.iec61970.base.auxiliaryequipment.FaultIndicator_pb2 import FaultIndicator as PBFaultIndicator
from zepben.protobuf.cim.iec61970.base.core.AcDcTerminal_pb2 import AcDcTerminal as PBAcDcTerminal
from zepben.protobuf.cim.iec61970.base.core.BaseVoltage_pb2 import BaseVoltage as PBBaseVoltage
from zepben.protobuf.cim.iec61970.base.core.ConductingEquipment_pb2 import ConductingEquipment as PBConductingEquipment
from zepben.protobuf.cim.iec61970.base.core.ConnectivityNodeContainer_pb2 import ConnectivityNodeContainer as PBConnectivityNodeContainer
from zepben.protobuf.cim.iec61970.base.core.ConnectivityNode_pb2 import ConnectivityNode as PBConnectivityNode
from zepben.protobuf.cim.iec61970.base.core.EquipmentContainer_pb2 import EquipmentContainer as PBEquipmentContainer
from zepben.protobuf.cim.iec61970.base.core.Equipment_pb2 import Equipment as PBEquipment
from zepben.protobuf.cim.iec61970.base.core.Feeder_pb2 import Feeder as PBFeeder
from zepben.protobuf.cim.iec61970.base.core.GeographicalRegion_pb2 import GeographicalRegion as PBGeographicalRegion
from zepben.protobuf.cim.iec61970.base.core.PhaseCode_pb2 import PhaseCode as PBPhaseCode
from zepben.protobuf.cim.iec61970.base.core.PowerSystemResource_pb2 import PowerSystemResource as PBPowerSystemResource
from zepben.protobuf.cim.iec61970.base.core.Site_pb2 import Site as PBSite
from zepben.protobuf.cim.iec61970.base.core.SubGeographicalRegion_pb2 import SubGeographicalRegion as PBSubGeographicalRegion
from zepben.protobuf.cim.iec61970.base.core.Substation_pb2 import Substation as PBSubstation
from zepben.protobuf.cim.iec61970.base.core.Terminal_pb2 import Terminal as PBTerminal
from zepben.protobuf.cim.iec61970.base.domain.UnitSymbol_pb2 import UnitSymbol as PBUnitSymbol
from zepben.protobuf.cim.iec61970.base.meas.Accumulator_pb2 import Accumulator as PBAccumulator
from zepben.protobuf.cim.iec61970.base.meas.Analog_pb2 import Analog as PBAnalog
from zepben.protobuf.cim.iec61970.base.meas.Control_pb2 import Control as PBControl
from zepben.protobuf.cim.iec61970.base.meas.Discrete_pb2 import Discrete as PBDiscrete
from zepben.protobuf.cim.iec61970.base.meas.IoPoint_pb2 import IoPoint as PBIoPoint
from zepben.protobuf.cim.iec61970.base.meas.Measurement_pb2 import Measurement as PBMeasurement
from zepben.protobuf.cim.iec61970.base.scada.RemoteControl_pb2 import RemoteControl as PBRemoteControl
from zepben.protobuf.cim.iec61970.base.scada.RemoteSource_pb2 import RemoteSource as PBRemoteSource
from zepben.protobuf.cim.iec61970.base.scada.RemotePoint_pb2 import RemotePoint as PBRemotePoint
from zepben.protobuf.cim.iec61970.base.wires.AcLineSegment_pb2 import AcLineSegment as PBAcLineSegment
from zepben.protobuf.cim.iec61970.base.wires.Breaker_pb2 import Breaker as PBBreaker
from zepben.protobuf.cim.iec61970.base.wires.Conductor_pb2 import Conductor as PBConductor
from zepben.protobuf.cim.iec61970.base.wires.Connector_pb2 import Connector as PBConnector
from zepben.protobuf.cim.iec61970.base.wires.Disconnector_pb2 import Disconnector as PBDisconnector
from zepben.protobuf.cim.iec61970.base.wires.EnergyConnection_pb2 import EnergyConnection as PBEnergyConnection
from zepben.protobuf.cim.iec61970.base.wires.EnergyConsumerPhase_pb2 import EnergyConsumerPhase as PBEnergyConsumerPhase
from zepben.protobuf.cim.iec61970.base.wires.EnergyConsumer_pb2 import EnergyConsumer as PBEnergyConsumer
from zepben.protobuf.cim.iec61970.base.wires.EnergySourcePhase_pb2 import EnergySourcePhase as PBEnergySourcePhase
from zepben.protobuf.cim.iec61970.base.wires.EnergySource_pb2 import EnergySource as PBEnergySource
from zepben.protobuf.cim.iec61970.base.wires.Fuse_pb2 import Fuse as PBFuse
from zepben.protobuf.cim.iec61970.base.wires.Jumper_pb2 import Jumper as PBJumper
from zepben.protobuf.cim.iec61970.base.wires.Junction_pb2 import Junction as PBJunction
from zepben.protobuf.cim.iec61970.base.wires.Line_pb2 import Line as PBLine
from zepben.protobuf.cim.iec61970.base.wires.LinearShuntCompensator_pb2 import LinearShuntCompensator as PBLinearShuntCompensator
from zepben.protobuf.cim.iec61970.base.wires.PerLengthImpedance_pb2 import PerLengthImpedance as PBPerLengthImpedance
from zepben.protobuf.cim.iec61970.base.wires.PerLengthLineParameter_pb2 import PerLengthLineParameter as PBPerLengthLineParameter
from zepben.protobuf.cim.iec61970.base.wires.PerLengthSequenceImpedance_pb2 import PerLengthSequenceImpedance as PBPerLengthSequenceImpedance
from zepben.protobuf.cim.iec61970.base.wires.PhaseShuntConnectionKind_pb2 import PhaseShuntConnectionKind as PBPhaseShuntConnectionKind
from zepben.protobuf.cim.iec61970.base.wires.PowerTransformerEnd_pb2 import PowerTransformerEnd as PBPowerTransformerEnd
from zepben.protobuf.cim.iec61970.base.wires.PowerTransformer_pb2 import PowerTransformer as PBPowerTransformer
from zepben.protobuf.cim.iec61970.base.wires.ProtectedSwitch_pb2 import ProtectedSwitch as PBProtectedSwitch
from zepben.protobuf.cim.iec61970.base.wires.RatioTapChanger_pb2 import RatioTapChanger as PBRatioTapChanger
from zepben.protobuf.cim.iec61970.base.wires.Recloser_pb2 import Recloser as PBRecloser
from zepben.protobuf.cim.iec61970.base.wires.RegulatingCondEq_pb2 import RegulatingCondEq as PBRegulatingCondEq
from zepben.protobuf.cim.iec61970.base.wires.ShuntCompensator_pb2 import ShuntCompensator as PBShuntCompensator
from zepben.protobuf.cim.iec61970.base.wires.SinglePhaseKind_pb2 import SinglePhaseKind as PBSinglePhaseKind
from zepben.protobuf.cim.iec61970.base.wires.Switch_pb2 import Switch as PBSwitch
from zepben.protobuf.cim.iec61970.base.wires.TapChanger_pb2 import TapChanger as PBTapChanger
from zepben.protobuf.cim.iec61970.base.wires.TransformerEnd_pb2 import TransformerEnd as PBTransformerEnd
from zepben.protobuf.cim.iec61970.base.wires.VectorGroup_pb2 import VectorGroup as PBVectorGroup
from zepben.protobuf.cim.iec61970.base.wires.WindingConnection_pb2 import WindingConnection as PBWindingConnection
from zepben.protobuf.cim.iec61970.infiec61970.feeder.Loop_pb2 import Loop as PBLoop
from zepben.protobuf.cim.iec61970.infiec61970.feeder.Circuit_pb2 import Circuit as PBCircuit
from zepben.protobuf.network.model.TracedPhases_pb2 import TracedPhases as PBTracedPhases

__all__ = ["CimTranslationException", "cableinfo_to_pb", "overheadwireinfo_to_pb", "wireinfo_to_pb", "asset_to_pb",
           "assetcontainer_to_pb", "assetinfo_to_pb",
           "assetorganisationrole_to_pb", "assetowner_to_pb", "pole_to_pb", "streetlight_to_pb", "structure_to_pb",
           "positionpoint_to_pb", "towndetail_to_pb", "streetaddress_to_pb",
           "location_to_pb", "enddevice_to_pb", "meter_to_pb", "usagepoint_to_pb", "operationalrestriction_to_pb",
           "auxiliaryequipment_to_pb", "faultindicator_to_pb", "acdcterminal_to_pb", "basevoltage_to_pb",
           "conductingequipment_to_pb", "connectivitynode_to_pb", "connectivitynodecontainer_to_pb", "equipment_to_pb",
           "equipmentcontainer_to_pb", "feeder_to_pb", "geographicalregion_to_pb", "powersystemresource_to_pb", "site_to_pb",
           "subgeographicalregion_to_pb", "substation_to_pb", "terminal_to_pb", "perlengthlineparameter_to_pb",
           "perlengthimpedance_to_pb", "aclinesegment_to_pb", "breaker_to_pb", "conductor_to_pb", "connector_to_pb",
           "disconnector_to_pb", "energyconnection_to_pb", "energyconsumer_to_pb", "energyconsumerphase_to_pb",
           "energysource_to_pb", "energysourcephase_to_pb", "fuse_to_pb", "jumper_to_pb", "junction_to_pb",
           "linearshuntcompensator_to_pb", "perlengthsequenceimpedance_to_pb", "powertransformer_to_pb",
           "powertransformerend_to_pb", "protectedswitch_to_pb", "ratiotapchanger_to_pb", "recloser_to_pb",
           "regulatingcondeq_to_pb", "shuntcompensator_to_pb", "switch_to_pb", "tapchanger_to_pb", "transformerend_to_pb",
           "tracedphases_to_pb"]


def get_or_none(getter, obj) -> object:
    return getter(obj) if obj else None


class CimTranslationException(Exception):
    pass


# IEC61968 ASSET INFO #
def cableinfo_to_pb(cim: CableInfo) -> PBCableInfo:
    return PBCableInfo(wi=wireinfo_to_pb(cim))


def overheadwireinfo_to_pb(cim: OverheadWireInfo) -> PBOverheadWireInfo:
    return PBOverheadWireInfo(wi=wireinfo_to_pb(cim))


def wireinfo_to_pb(cim: WireInfo) -> PBWireInfo:
    return PBWireInfo(ai=assetinfo_to_pb(cim),
                      ratedCurrent=cim.rated_current,
                      material=PBWireMaterialKind.Value(cim.material.short_name))


# IEC61968 ASSETS #
def asset_to_pb(cim: Asset) -> PBAsset:
    return PBAsset(io=identifiedobject_to_pb(cim),
                   locationMRID=cim.location.mrid if cim.location else None,
                   organisationRoleMRIDs=[str(io.mrid) for io in cim.organisation_roles])


def assetcontainer_to_pb(cim: AssetContainer) -> PBAssetContainer:
    return PBAssetContainer(at=asset_to_pb(cim))


def assetinfo_to_pb(cim: AssetInfo) -> PBAssetInfo:
    return PBAssetInfo(io=identifiedobject_to_pb(cim))


def assetorganisationrole_to_pb(cim: AssetOrganisationRole) -> PBAssetOrganisationRole:
    pb = PBAssetOrganisationRole()
    getattr(pb, "or").CopyFrom(organisationrole_to_pb(cim))
    return pb


def assetowner_to_pb(cim: AssetOwner) -> PBAssetOwner:
    return PBAssetOwner(aor=assetorganisationrole_to_pb(cim))


def pole_to_pb(cim: Pole) -> PBPole:
    return PBPole(st=structure_to_pb(cim), streetlightMRIDs=[str(io.mrid) for io in cim.streetlights], classification=cim.classification)


def streetlight_to_pb(cim: Streetlight) -> PBStreetlight:
    return PBStreetlight(at=asset_to_pb(cim),
                         poleMRID=str(cim.pole.mrid),
                         lightRating=cim.light_rating,
                         lampKind=PBStreetlightLampKind.Value(cim.lamp_kind.short_name))


def structure_to_pb(cim: Structure) -> PBStructure:
    return PBStructure(ac=assetcontainer_to_pb(cim))


# IEC61968 COMMON #
def location_to_pb(cim: Location) -> PBLocation:
    return PBLocation(io=identifiedobject_to_pb(cim),
                      mainAddress=get_or_none(streetaddress_to_pb, cim.main_address),
                      positionPoints=[positionpoint_to_pb(point) for _, point in cim.points])


def positionpoint_to_pb(cim: PositionPoint) -> PBPositionPoint:
    return PBPositionPoint(xPosition=cim.x_position, yPosition=cim.y_position)


def streetaddress_to_pb(cim: StreetAddress) -> PBStreetAddress:
    return PBStreetAddress(postalCode=cim.postal_code, townDetail=get_or_none(towndetail_to_pb, cim.town_detail))


def towndetail_to_pb(cim: TownDetail) -> PBTownDetail:
    return PBTownDetail(name=cim.name, stateOrProvince=cim.state_or_province)


# IEC61968 METERING #
def enddevice_to_pb(cim: EndDevice) -> PBEndDevice:
    return PBEndDevice(ac=assetcontainer_to_pb(cim),
                       usagePointMRIDs=[str(io.mrid) for io in cim.usage_points],
                       customerMRID=cim.customer_mrid,
                       serviceLocationMRID=mrid_or_empty(cim.service_location))


def meter_to_pb(cim: Meter) -> PBMeter:
    return PBMeter(ed=enddevice_to_pb(cim))


def usagepoint_to_pb(cim: UsagePoint) -> PBUsagePoint:
    return PBUsagePoint(io=identifiedobject_to_pb(cim),
                        usagePointLocationMRID=mrid_or_empty(cim.usage_point_location),
                        equipmentMRIDs=[str(io.mrid) for io in cim.equipment],
                        endDeviceMRIDs=[str(io.mrid) for io in cim.end_devices])


# IEC61968 OPERATIONS #
def operationalrestriction_to_pb(cim: OperationalRestriction) -> PBOperationalRestriction:
    return PBOperationalRestriction(doc=document_to_pb(cim),
                                    equipmentMRIDs=[str(io.mrid) for io in cim.equipment])


# IEC61970 AUXILIARY EQUIPMENT #
def auxiliaryequipment_to_pb(cim: AuxiliaryEquipment) -> PBAuxiliaryEquipment:
    return PBAuxiliaryEquipment(eq=equipment_to_pb(cim),
                                terminalMRID=mrid_or_empty(cim.terminal))


def faultindicator_to_pb(cim: FaultIndicator) -> PBFaultIndicator:
    return PBFaultIndicator(ae=auxiliaryequipment_to_pb(cim))


# IEC61970 CORE #
def acdcterminal_to_pb(cim: AcDcTerminal) -> PBAcDcTerminal:
    return PBAcDcTerminal(io=identifiedobject_to_pb(cim))


def basevoltage_to_pb(cim: BaseVoltage) -> PBBaseVoltage:
    return PBBaseVoltage(io=identifiedobject_to_pb(cim),
                         nominalVoltage=cim.nominal_voltage)


def conductingequipment_to_pb(cim: ConductingEquipment) -> PBConductingEquipment:
    return PBConductingEquipment(eq=equipment_to_pb(cim),
                                 baseVoltageMRID=mrid_or_empty(cim.base_voltage),
                                 terminalMRIDs=[str(io.mrid) for io in cim.terminals])


def connectivitynode_to_pb(cim: ConnectivityNode) -> PBConnectivityNode:
    return PBConnectivityNode(io=identifiedobject_to_pb(cim), terminalMRIDs=[str(io.mrid) for io in cim.terminals])


def connectivitynodecontainer_to_pb(cim: ConnectivityNodeContainer) -> PBConnectivityNodeContainer:
    return PBConnectivityNodeContainer(psr=powersystemresource_to_pb(cim))


def equipment_to_pb(cim: Equipment) -> PBEquipment:
    pb = PBEquipment(psr=powersystemresource_to_pb(cim),
                     inService=cim.in_service,
                     normallyInService=cim.normally_in_service,
                     equipmentContainerMRIDs=[str(io.mrid) for io in cim.equipment_containers],
                     usagePointMRIDs=[str(io.mrid) for io in cim.usage_points],
                     operationalRestrictionMRIDs=[str(io.mrid) for io in cim.operational_restrictions],
                     currentFeederMRIDs=[str(io.mrid) for io in cim.current_feeders])
    return pb


def equipmentcontainer_to_pb(cim: EquipmentContainer) -> PBEquipmentContainer:
    return PBEquipmentContainer(cnc=connectivitynodecontainer_to_pb(cim),
                                equipmentMRIDs=[str(io.mrid) for io in cim.equipment])


def feeder_to_pb(cim: Feeder) -> PBFeeder:
    return PBFeeder(ec=equipmentcontainer_to_pb(cim),
                    normalHeadTerminalMRID=mrid_or_empty(cim.normal_head_terminal),
                    normalEnergizingSubstationMRID=mrid_or_empty(cim.normal_energizing_substation),
                    currentEquipmentMRIDs=[str(io.mrid) for io in cim.current_equipment])


def geographicalregion_to_pb(cim: GeographicalRegion) -> PBGeographicalRegion:
    return PBGeographicalRegion(io=identifiedobject_to_pb(cim),
                                subGeographicalRegionMRIDs=[str(io.mrid) for io in cim.sub_geographical_regions])


def powersystemresource_to_pb(cim: PowerSystemResource) -> PBPowerSystemResource:
    return PBPowerSystemResource(io=identifiedobject_to_pb(cim),
                                 assetInfoMRID=mrid_or_empty(cim.asset_info),
                                 locationMRID=mrid_or_empty(cim.location))


def site_to_pb(cim: Site) -> PBSite:
    return PBSite(ec=equipmentcontainer_to_pb(cim))


def subgeographicalregion_to_pb(cim: SubGeographicalRegion) -> PBSubGeographicalRegion:
    return PBSubGeographicalRegion(io=identifiedobject_to_pb(cim),
                                   geographicalRegionMRID=mrid_or_empty(cim.geographical_region),
                                   substationMRIDs=[str(io.mrid) for io in cim.substations])


def substation_to_pb(cim: Substation) -> PBSubstation:
    return PBSubstation(ec=equipmentcontainer_to_pb(cim),
                        subGeographicalRegionMRID=mrid_or_empty(cim.sub_geographical_region),
                        normalEnergizedFeederMRIDs=[str(io.mrid) for io in cim.feeders],
                        loopMRIDs=[str(io.mrid) for io in cim.loops],
                        normalEnergizedLoopMRIDs=[str(io.mrid) for io in cim.energized_loops],
                        circuitMRIDs=[str(io.mrid) for io in cim.circuits])


def terminal_to_pb(cim: Terminal) -> PBTerminal:
    return PBTerminal(ad=acdcterminal_to_pb(cim),
                      conductingEquipmentMRID=mrid_or_empty(cim.conducting_equipment),
                      connectivityNodeMRID=mrid_or_empty(cim.connectivity_node),
                      tracedPhases=get_or_none(tracedphases_to_pb, cim.traced_phases),
                      phases=PBPhaseCode.Value(cim.phases.short_name),
                      sequenceNumber=cim.sequence_number)


# IEC61970 WIRES #
def aclinesegment_to_pb(cim: AcLineSegment) -> PBAcLineSegment:
    return PBAcLineSegment(cd=conductor_to_pb(cim), perLengthSequenceImpedanceMRID=mrid_or_empty(cim.per_length_sequence_impedance))


def breaker_to_pb(cim: Breaker) -> PBBreaker:
    return PBBreaker(sw=protectedswitch_to_pb(cim))


def conductor_to_pb(cim: Conductor) -> PBConductor:
    return PBConductor(ce=conductingequipment_to_pb(cim), length=cim.length)


def connector_to_pb(cim: Connector) -> PBConnector:
    return PBConnector(ce=conductingequipment_to_pb(cim))


def disconnector_to_pb(cim: Disconnector) -> PBDisconnector:
    return PBDisconnector(sw=switch_to_pb(cim))


def energyconnection_to_pb(cim: EnergyConnection) -> PBEnergyConnection:
    return PBEnergyConnection(ce=conductingequipment_to_pb(cim))


def energyconsumer_to_pb(cim: EnergyConsumer) -> PBEnergyConsumer:
    return PBEnergyConsumer(ec=energyconnection_to_pb(cim),
                            energyConsumerPhasesMRIDs=[str(io.mrid) for io in cim.phases],
                            customerCount=cim.customer_count,
                            grounded=cim.grounded,
                            p=cim.p,
                            pFixed=cim.p_fixed,
                            phaseConnection=PBPhaseShuntConnectionKind.Enum.Value(cim.phase_connection.short_name),
                            q=cim.q,
                            qFixed=cim.q_fixed)


def energyconsumerphase_to_pb(cim: EnergyConsumerPhase) -> PBEnergyConsumerPhase:
    return PBEnergyConsumerPhase(psr=powersystemresource_to_pb(cim),
                                 energyConsumerMRID=mrid_or_empty(cim.energy_consumer),
                                 phase=PBSinglePhaseKind.Value(cim.phase.short_name),
                                 p=cim.p,
                                 pFixed=cim.p_fixed,
                                 q=cim.q,
                                 qFixed=cim.q_fixed)


def energysource_to_pb(cim: EnergySource) -> PBEnergySource:
    return PBEnergySource(ec=energyconnection_to_pb(cim),
                          energySourcePhasesMRIDs=[str(io.mrid) for io in cim.phases],
                          activePower=cim.active_power,
                          reactivePower=cim.reactive_power,
                          voltageAngle=cim.voltage_angle,
                          voltageMagnitude=cim.voltage_magnitude,
                          r=cim.r,
                          x=cim.x,
                          pMax=cim.p_max,
                          pMin=cim.p_min,
                          r0=cim.r0,
                          rn=cim.rn,
                          x0=cim.x0,
                          xn=cim.xn)


def energysourcephase_to_pb(cim: EnergySourcePhase) -> PBEnergySourcePhase:
    return PBEnergySourcePhase(psr=powersystemresource_to_pb(cim),
                               energySourceMRID=mrid_or_empty(cim.energy_source),
                               phase=PBSinglePhaseKind.Value(cim.phase.short_name))


def fuse_to_pb(cim: Fuse) -> PBFuse:
    return PBFuse(sw=switch_to_pb(cim))


def jumper_to_pb(cim: Jumper) -> PBJumper:
    return PBJumper(sw=switch_to_pb(cim))


def junction_to_pb(cim: Junction) -> PBJunction:
    return PBJunction(cn=connector_to_pb(cim))


def line_to_pb(cim: Line) -> PBLine:
    return PBLine(ec=equipmentcontainer_to_pb(cim))


def linearshuntcompensator_to_pb(cim: LinearShuntCompensator) -> PBLinearShuntCompensator:
    return PBLinearShuntCompensator(sc=shuntcompensator_to_pb(cim),
                                    b0PerSection=cim.b0_per_section,
                                    bPerSection=cim.b_per_section,
                                    g0PerSection=cim.g0_per_section,
                                    gPerSection=cim.g_per_section)


def perlengthlineparameter_to_pb(cim: PerLengthLineParameter) -> PBPerLengthLineParameter:
    return PBPerLengthLineParameter(io=identifiedobject_to_pb(cim))


def perlengthimpedance_to_pb(cim: PerLengthImpedance) -> PBPerLengthImpedance:
    return PBPerLengthImpedance(lp=perlengthlineparameter_to_pb(cim))


def perlengthsequenceimpedance_to_pb(cim: PerLengthSequenceImpedance) -> PBPerLengthSequenceImpedance:
    return PBPerLengthSequenceImpedance(pli=perlengthimpedance_to_pb(cim),
                                        r=cim.r,
                                        x=cim.x,
                                        r0=cim.r0,
                                        x0=cim.x0,
                                        bch=cim.bch,
                                        gch=cim.gch,
                                        b0ch=cim.b0ch,
                                        g0ch=cim.g0ch)


def powertransformer_to_pb(cim: PowerTransformer) -> PBPowerTransformer:
    return PBPowerTransformer(ce=conductingequipment_to_pb(cim),
                              powerTransformerEndMRIDs=[str(io.mrid) for io in cim.ends],
                              vectorGroup=PBVectorGroup.Value(cim.vector_group.short_name))


def powertransformerend_to_pb(cim: PowerTransformerEnd) -> PBPowerTransformerEnd:
    return PBPowerTransformerEnd(te=transformerend_to_pb(cim),
                                 powerTransformerMRID=mrid_or_empty(cim.power_transformer),
                                 ratedS=cim.rated_s,
                                 ratedU=cim.rated_u,
                                 r=cim.r,
                                 r0=cim.r0,
                                 x=cim.x,
                                 x0=cim.x0,
                                 connectionKind=PBWindingConnection.Value(cim.connection_kind.short_name),
                                 b=cim.b,
                                 b0=cim.b0,
                                 g=cim.g,
                                 g0=cim.g0,
                                 phaseAngleClock=cim.phase_angle_clock)


def protectedswitch_to_pb(cim: ProtectedSwitch) -> PBProtectedSwitch:
    return PBProtectedSwitch(sw=switch_to_pb(cim))


def ratiotapchanger_to_pb(cim: RatioTapChanger) -> PBRatioTapChanger:
    return PBRatioTapChanger(tc=tapchanger_to_pb(cim),
                             transformerEndMRID=mrid_or_empty(cim.transformer_end),
                             stepVoltageIncrement=cim.step_voltage_increment)


def recloser_to_pb(cim: Recloser) -> PBRecloser:
    return PBRecloser(sw=protectedswitch_to_pb(cim))


def regulatingcondeq_to_pb(cim: RegulatingCondEq) -> PBRegulatingCondEq:
    return PBRegulatingCondEq(ec=energyconnection_to_pb(cim), controlEnabled=cim.control_enabled)


def shuntcompensator_to_pb(cim: ShuntCompensator) -> PBShuntCompensator:
    return PBShuntCompensator(rce=regulatingcondeq_to_pb(cim),
                              sections=cim.sections,
                              grounded=cim.grounded,
                              nomU=cim.nom_u,
                              phaseConnection=PBPhaseShuntConnectionKind.Value(cim.phase_connection))


def switch_to_pb(cim: Switch) -> PBSwitch:
    return PBSwitch(ce=conductingequipment_to_pb(cim),
                    normalOpen=cim.get_normal_state(),
                    open=cim.get_state())


def tapchanger_to_pb(cim: TapChanger) -> PBTapChanger:
    return PBTapChanger(psr=powersystemresource_to_pb(cim),
                        highStep=cim.high_step,
                        lowStep=cim.low_step,
                        step=cim.step,
                        neutralStep=cim.neutral_step,
                        neutralU=cim.neutral_u,
                        normalStep=cim.normal_step,
                        controlEnabled=cim.control_enabled)


def transformerend_to_pb(cim: TransformerEnd) -> PBTransformerEnd:
    return PBTransformerEnd(io=identifiedobject_to_pb(cim),
                            terminalMRID=mrid_or_empty(cim.terminal),
                            baseVoltageMRID=mrid_or_empty(cim.base_voltage),
                            ratioTapChangerMRID=mrid_or_empty(cim.ratio_tap_changer),
                            endNumber=cim.end_number,
                            grounded=cim.grounded,
                            rGround=cim.r_ground,
                            xGround=cim.x_ground)


def circuit_to_pb(cim: Circuit) -> PBCircuit:
    return PBCircuit(l=line_to_pb(cim),
                     loopMRID=mrid_or_empty(cim.loop.mrid),
                     endTerminalMRIDs=[str(io.mrid) for io in cim.end_terminals],
                     endSubstationMRIDs=[str(io.mrid) for io in cim.end_substations])


def loop_to_pb(cim: Loop) -> PBLoop:
    return PBLoop(io=identifiedobject_to_pb(cim),
                  circuitMRIDs=[str(io.mrid) for io in cim.circuits],
                  substationMRIDs=[str(io.mrid) for io in cim.substations],
                  normalEnergizingSubstationMRIDs=[str(io.mrid) for io in cim.energizing_substations])


# IEC61970 MEAS #
def control_to_pb(cim: Control) -> PBControl:
    return PBControl(ip=iopoint_to_pb(cim),
                     remoteControlMRID=mrid_or_empty(cim.remote_control),
                     powerSystemResourceMRID=cim.power_system_resource_mrid)


def iopoint_to_pb(cim: IoPoint) -> PBIoPoint:
    return PBIoPoint(io=identifiedobject_to_pb(cim))


def accumulator_to_pb(cim: Accumulator) -> PBAccumulator:
    return PBAccumulator(measurement=measurement_to_pb(cim))


def analog_to_pb(cim: Analog) -> PBAnalog:
    return PBAnalog(measurement=measurement_to_pb(cim), positiveFlowIn=cim.positive_flow_in)


def discrete_to_pb(cim: Discrete) -> PBDiscrete:
    return PBDiscrete(measurement=measurement_to_pb(cim))


def measurement_to_pb(cim: Measurement) -> PBMeasurement:
    return PBMeasurement(io=identifiedobject_to_pb(cim),
                         remoteSourceMRID=mrid_or_empty(cim.remote_source),
                         powerSystemResourceMRID=cim.power_system_resource_mrid,
                         terminalMRID=cim.terminal_mrid,
                         phases=PBPhaseCode.Value(cim.phases.short_name),
                         unitSymbol=PBUnitSymbol.Value(cim.unitSymbol.short_name))


# IEC61970 SCADA #
def remotecontrol_to_pb(cim: RemoteControl) -> PBRemoteControl:
    return PBRemoteControl(rp=remotepoint_to_pb(cim), controlMRID=mrid_or_empty(cim.control))


def remotepoint_to_pb(cim: RemotePoint) -> PBRemotePoint:
    return PBRemotePoint(io=identifiedobject_to_pb(cim))


def remotesource_to_pb(cim: RemoteSource) -> PBRemoteSource:
    return PBRemoteSource(rp=remotepoint_to_pb(cim), measurementMRID=mrid_or_empty(cim.measurement))


# MODEL #
def tracedphases_to_pb(cim: TracedPhases) -> PBTracedPhases:
    return PBTracedPhases(normalStatus=cim._normal_status, currentStatus=cim._current_status)


# Extension functions for each CIM type.
CableInfo.to_pb = lambda self: cableinfo_to_pb(self)
OverheadWireInfo.to_pb = lambda self: overheadwireinfo_to_pb(self)
WireInfo.to_pb = lambda self: wireinfo_to_pb(self)
Asset.to_pb = lambda self: asset_to_pb(self)
AssetContainer.to_pb = lambda self: assetcontainer_to_pb(self)
AssetInfo.to_pb = lambda self: assetinfo_to_pb(self)
AssetOrganisationRole.to_pb = lambda self: assetorganisationrole_to_pb(self)
AssetOwner.to_pb = lambda self: assetowner_to_pb(self)
Pole.to_pb = lambda self: pole_to_pb(self)
Streetlight.to_pb = lambda self: streetlight_to_pb(self)
Structure.to_pb = lambda self: structure_to_pb(self)
PositionPoint.to_pb = lambda self: positionpoint_to_pb(self)
TownDetail.to_pb = lambda self: towndetail_to_pb(self)
StreetAddress.to_pb = lambda self: streetaddress_to_pb(self)
Location.to_pb = lambda self: location_to_pb(self)
EndDevice.to_pb = lambda self: enddevice_to_pb(self)
Meter.to_pb = lambda self: meter_to_pb(self)
UsagePoint.to_pb = lambda self: usagepoint_to_pb(self)
OperationalRestriction.to_pb = lambda self: operationalrestriction_to_pb(self)
AuxiliaryEquipment.to_pb = lambda self: auxiliaryequipment_to_pb(self)
FaultIndicator.to_pb = lambda self: faultindicator_to_pb(self)
AcDcTerminal.to_pb = lambda self: acdcterminal_to_pb(self)
BaseVoltage.to_pb = lambda self: basevoltage_to_pb(self)
ConductingEquipment.to_pb = lambda self: conductingequipment_to_pb(self)
ConnectivityNode.to_pb = lambda self: connectivitynode_to_pb(self)
ConnectivityNodeContainer.to_pb = lambda self: connectivitynodecontainer_to_pb(self)
Equipment.to_pb = lambda self: equipment_to_pb(self)
EquipmentContainer.to_pb = lambda self: equipmentcontainer_to_pb(self)
Feeder.to_pb = lambda self: feeder_to_pb(self)
GeographicalRegion.to_pb = lambda self: geographicalregion_to_pb(self)
PowerSystemResource.to_pb = lambda self: powersystemresource_to_pb(self)
Site.to_pb = lambda self: site_to_pb(self)
SubGeographicalRegion.to_pb = lambda self: subgeographicalregion_to_pb(self)
Substation.to_pb = lambda self: substation_to_pb(self)
Terminal.to_pb = lambda self: terminal_to_pb(self)
PerLengthLineParameter.to_pb = lambda self: perlengthlineparameter_to_pb(self)
PerLengthImpedance.to_pb = lambda self: perlengthimpedance_to_pb(self)
AcLineSegment.to_pb = lambda self: aclinesegment_to_pb(self)
Breaker.to_pb = lambda self: breaker_to_pb(self)
Conductor.to_pb = lambda self: conductor_to_pb(self)
Connector.to_pb = lambda self: connector_to_pb(self)
Disconnector.to_pb = lambda self: disconnector_to_pb(self)
EnergyConnection.to_pb = lambda self: energyconnection_to_pb(self)
EnergyConsumer.to_pb = lambda self: energyconsumer_to_pb(self)
EnergyConsumerPhase.to_pb = lambda self: energyconsumerphase_to_pb(self)
EnergySource.to_pb = lambda self: energysource_to_pb(self)
EnergySourcePhase.to_pb = lambda self: energysourcephase_to_pb(self)
Fuse.to_pb = lambda self: fuse_to_pb(self)
Jumper.to_pb = lambda self: jumper_to_pb(self)
Junction.to_pb = lambda self: junction_to_pb(self)
Line.to_pb = line_to_pb
LinearShuntCompensator.to_pb = lambda self: linearshuntcompensator_to_pb(self)
PerLengthSequenceImpedance.to_pb = lambda self: perlengthsequenceimpedance_to_pb(self)
PowerTransformer.to_pb = lambda self: powertransformer_to_pb(self)
PowerTransformerEnd.to_pb = lambda self: powertransformerend_to_pb(self)
ProtectedSwitch.to_pb = lambda self: protectedswitch_to_pb(self)
RatioTapChanger.to_pb = lambda self: ratiotapchanger_to_pb(self)
Recloser.to_pb = lambda self: recloser_to_pb(self)
RegulatingCondEq.to_pb = lambda self: regulatingcondeq_to_pb(self)
ShuntCompensator.to_pb = lambda self: shuntcompensator_to_pb(self)
Switch.to_pb = lambda self: switch_to_pb(self)
TapChanger.to_pb = lambda self: tapchanger_to_pb(self)
TransformerEnd.to_pb = lambda self: transformerend_to_pb(self)
Circuit.to_pb = circuit_to_pb
Loop.to_pb = loop_to_pb
Control.to_pb = control_to_pb
IoPoint.to_pb = iopoint_to_pb
Accumulator.to_pb = accumulator_to_pb
Analog.to_pb = analog_to_pb
Discrete.to_pb = discrete_to_pb
Measurement.to_pb = measurement_to_pb
RemoteControl.to_pb = remotecontrol_to_pb
RemotePoint.to_pb = remotepoint_to_pb
RemoteSource.to_pb = remotesource_to_pb
TracedPhases.to_pb = tracedphases_to_pb



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.protobuf.cim.iec61970.base.diagramlayout.Diagram_pb2 import Diagram as PBDiagram
from zepben.protobuf.cim.iec61970.base.diagramlayout.OrientationKind_pb2 import OrientationKind as PBOrientationKind
from zepben.protobuf.cim.iec61970.base.diagramlayout.DiagramStyle_pb2 import DiagramStyle as PBDiagramStyle
from zepben.protobuf.cim.iec61970.base.diagramlayout.DiagramObject_pb2 import DiagramObject as PBDiagramObject
from zepben.protobuf.cim.iec61970.base.diagramlayout.DiagramObjectPoint_pb2 import DiagramObjectPoint as PBDiagramObjectPoint
from zepben.protobuf.cim.iec61970.base.diagramlayout.DiagramObjectStyle_pb2 import DiagramObjectStyle as PBDiagramObjectStyle

from zepben.evolve.services.common.translator.util import mrid_or_empty
from zepben.evolve.model import Diagram, DiagramObject, DiagramObjectPoint
from zepben.evolve.services.common.translator.base_cim2proto import identifiedobject_to_pb

__all__ = ["diagram_to_pb", "diagramobject_to_pb", "diagramobjectpoint_to_pb"]


# IEC61970 DIAGRAMLAYOUT #
def diagram_to_pb(cim: Diagram) -> PBDiagram:
    return PBDiagram(io=identifiedobject_to_pb(cim),
                     diagramStyle=PBDiagramStyle.Value(cim.diagram_style.short_name),
                     orientationKind=PBOrientationKind.Value(cim.orientation_kind.short_name),
                     diagramObjectMRIDs=[str(io.mrid) for io in cim.diagram_objects])


def diagramobject_to_pb(cim: DiagramObject) -> PBDiagramObject:
    return PBDiagramObject(io=identifiedobject_to_pb(cim),
                           diagramMRID=mrid_or_empty(cim.diagram),
                           identifiedObjectMRID=cim.identified_object_mrid,
                           diagramObjectStyle=PBDiagramObjectStyle.Value(cim.style.short_name),
                           rotation=cim.rotation,
                           diagramObjectPoints=[diagramobjectpoint_to_pb(io) for io in cim.points])


def diagramobjectpoint_to_pb(cim: DiagramObjectPoint) -> PBDiagramObjectPoint:
    return PBDiagramObjectPoint(xPosition=cim.x_position, yPosition=cim.y_position)


Diagram.to_pb = diagram_to_pb
DiagramObject.to_pb = diagramobject_to_pb
DiagramObjectPoint.to_pb = diagramobjectpoint_to_pb



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations
from typing import Optional, Dict

from dataclassy import dataclass

__all__ = ["NetworkHierarchy", "NetworkHierarchyIdentifiedObject", "NetworkHierarchyFeeder", "NetworkHierarchyGeographicalRegion", "NetworkHierarchySubstation",
           "NetworkHierarchySubGeographicalRegion"]


@dataclass(slots=True)
class NetworkHierarchy(object):
    """Container for simplified network hierarchy objects"""
    geographical_regions: Dict[str, NetworkHierarchyGeographicalRegion]
    sub_geographical_regions: Dict[str, NetworkHierarchySubGeographicalRegion]
    substations: Dict[str, NetworkHierarchySubstation]
    feeders: Dict[str, NetworkHierarchyFeeder]


@dataclass(slots=True)
class NetworkHierarchyIdentifiedObject(object):
    """A simplified representation of an identified object for requesting the network hierarchy."""
    mrid: str
    name: str


class NetworkHierarchyGeographicalRegion(NetworkHierarchyIdentifiedObject):
    """A simplified representation of a geographical region for requesting the network hierarchy."""
    sub_geographical_regions: Dict[str, NetworkHierarchySubGeographicalRegion] = None


class NetworkHierarchySubGeographicalRegion(NetworkHierarchyIdentifiedObject):
    """A simplified representation of a subgeographical region for requesting the network hierarchy."""
    substations: Dict[str, NetworkHierarchySubstation] = None
    geographical_region: Optional[NetworkHierarchyGeographicalRegion] = None


class NetworkHierarchySubstation(NetworkHierarchyIdentifiedObject):
    """A simplified representation of a substation for requesting the network hierarchy."""
    feeders: Dict[str, NetworkHierarchyFeeder]
    sub_geographical_region: Optional[NetworkHierarchySubGeographicalRegion] = None


class NetworkHierarchyFeeder(NetworkHierarchyIdentifiedObject):
    """A simplified representation of a feeder for requesting the network hierarchy."""
    substation: Optional[NetworkHierarchySubstation] = None


#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.protobuf.cim.iec61968.common.Agreement_pb2 import Agreement as PBAgreement
from zepben.protobuf.cim.iec61968.customers.CustomerAgreement_pb2 import CustomerAgreement as PBCustomerAgreement
from zepben.protobuf.cim.iec61968.customers.CustomerKind_pb2 import CustomerKind as PBCustomerKind
from zepben.protobuf.cim.iec61968.customers.Customer_pb2 import Customer as PBCustomer
from zepben.protobuf.cim.iec61968.customers.PricingStructure_pb2 import PricingStructure as PBPricingStructure
from zepben.protobuf.cim.iec61968.customers.Tariff_pb2 import Tariff as PBTariff

from zepben.evolve.services.common.translator.util import mrid_or_empty
from zepben.evolve.model import Customer
from zepben.evolve.services.common.translator.base_cim2proto import document_to_pb, organisation_to_pb
from zepben.evolve.model import Agreement
from zepben.evolve.model import CustomerAgreement, PricingStructure, Tariff

__all__ = ["agreement_to_pb", "customer_to_pb", "customeragreement_to_pb", "pricingstructure_to_pb", "tariff_to_pb"]


def agreement_to_pb(cim: Agreement) -> PBAgreement:
    return PBAgreement(doc=document_to_pb(cim))


def customer_to_pb(cim: Customer) -> PBCustomer:
    cust = PBCustomer(kind=PBCustomerKind.Value(cim.kind.short_name), agreementMRIDs=[str(io.mrid) for io in cim.agreements])
    setattr(cust, 'or', organisation_to_pb(cim))
    return cust


def customeragreement_to_pb(cim: CustomerAgreement) -> PBCustomerAgreement:
    return PBCustomerAgreement(agr=agreement_to_pb(cim),
                               customerMRID=mrid_or_empty(cim.customer),
                               pricingStructureMRIDs=[str(io.mrid) for io in cim.pricing_structures])


def pricingstructure_to_pb(cim: PricingStructure) -> PBPricingStructure:
    return PBPricingStructure(doc=document_to_pb(cim), tariffMRIDs=[str(io.mrid) for io in cim.tariffs])


def tariff_to_pb(cim: Tariff) -> PBTariff:
    return PBTariff(doc=document_to_pb(cim))


Agreement.to_pb = agreement_to_pb
Customer.to_pb = customer_to_pb
CustomerAgreement.to_pb = customeragreement_to_pb
PricingStructure.to_pb = pricingstructure_to_pb
Tariff.to_pb = tariff_to_pb




#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import List, Optional, Generator

from zepben.evolve.model.cim.iec61970.base.core.base_voltage import BaseVoltage
from zepben.evolve.model.cim.iec61970.base.core.equipment import Equipment

__all__ = ['ConductingEquipment']

from zepben.evolve.util import get_by_mrid, require


class ConductingEquipment(Equipment):
    """
    Abstract class, should only be used through subclasses.
    The parts of the AC power system that are designed to carry current or that are conductively connected through
    terminals.

    ConductingEquipment are connected by `zepben.evolve.cim.iec61970.base.core.Terminal`'s which are in turn associated with
    `zepben.evolve.cim.iec61970.base.connectivity_node.ConnectivityNode`'s. Each `zepben.evolve.iec61970.base.core.terminal.Terminal` is associated with
    _exactly one_ `ConnectivityNode`, and through that `ConnectivityNode` can be linked with many other `Terminals` and `ConductingEquipment`.
    """

    base_voltage: Optional[BaseVoltage] = None
    """`zepben.evolve.iec61970.base.core.base_voltage.BaseVoltage` of this `ConductingEquipment`. Use only when there is no voltage level container used and 
    only one base voltage applies. For example, not used for transformers."""

    _terminals: List[Terminal] = []

    def __init__(self, usage_points: List[UsagePoint] = None, equipment_containers: List[EquipmentContainer] = None,
                 operational_restrictions: List[OperationalRestriction] = None, current_feeders: List[Feeder] = None, terminals: List[Terminal] = None):
        super(ConductingEquipment, self).__init__(usage_points, equipment_containers, operational_restrictions, current_feeders)
        if terminals:
            for term in terminals:
                self.add_terminal(term)

    def get_base_voltage(self, terminal: Terminal = None):
        """
        Get the `zepben.evolve.iec61970.base.core.base_voltage.BaseVoltage` of this `ConductingEquipment`.
        Note `terminal` is not used here, but this method can be overridden in child classes (e.g PowerTransformer).

        :param terminal: The `zepben.evolve.cim.iec61970.base.core.terminal.Terminal` to get the voltage at.
        :return: The BaseVoltage of this `ConductingEquipment` at `terminal`
        """
        return self.base_voltage

    @property
    def terminals(self) -> Generator[Terminal, None, None]:
        """
        `ConductingEquipment` have `zepben.evolve.cim.iec61970.base.core.terminal.Terminal`s that may be connected to other `ConductingEquipment`
        `zepben.evolve.cim.iec61970.base.core.terminal.Terminal`s via `ConnectivityNode`s.
        """
        for term in self._terminals:
            yield term

    def num_terminals(self):
        """
        Get the number of `zepben.evolve.cim.iec61970.base.core.terminal.Terminal`s for this `ConductingEquipment`.
        """
        return len(self._terminals)

    def get_terminal_by_mrid(self, mrid: str) -> Terminal:
        """
        Get the `zepben.evolve.iec61970.base.core.terminal.Terminal` for this `ConductingEquipment` identified by `mrid`

        `mrid` the mRID of the required `zepben.evolve.cim.iec61970.base.core.terminal.Terminal`
        Returns The `zepben.evolve.cim.iec61970.base.core.terminal.Terminal` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._terminals, mrid)

    def get_terminal_by_sn(self, sequence_number: int):
        """
        Get the `zepben.evolve.iec61970.base.core.terminal.Terminal` on this `ConductingEquipment` by its `sequence_number`.

        `sequence_number` The `sequence_number` of the `zepben.evolve.iec61970.base.core.terminal.Terminal` in relation to this `ConductingEquipment`.
        Returns The `zepben.evolve.iec61970.base.core.terminal.Terminal` on this `ConductingEquipment` with sequence number `sequence_number`
        Raises IndexError if no `zepben.evolve.iec61970.base.core.terminal.Terminal` was found with sequence_number `sequence_number`.
        """
        for term in self._terminals:
            if term.sequence_number == sequence_number:
                return term
        raise IndexError(f"No Terminal with sequence_number {sequence_number} was found in ConductingEquipment {str(self)}")

    def __getitem__(self, item: int):
        return self.get_terminal_by_sn(item)

    def add_terminal(self, terminal: Terminal) -> ConductingEquipment:
        """
        Associate `terminal` with this `ConductingEquipment`. If `terminal.sequence_number` == 0, the terminal will be assigned a sequence_number of
        `self.num_terminals() + 1`.

        `terminal` The `zepben.evolve.cim.iec61970.base.core.terminal.Terminal` to associate with this `ConductingEquipment`.
        Returns A reference to this `ConductingEquipment` to allow fluent use.
        Raises `ValueError` if another `zepben.evolve.iec61970.base.core.terminal.Terminal` with the same `mrid` already exists for this `ConductingEquipment`.
        """
        if self._validate_terminal(terminal):
            return self

        if terminal.sequence_number == 0:
            terminal.sequence_number = self.num_terminals() + 1

        self._terminals.append(terminal)
        self._terminals.sort(key=lambda t: t.sequence_number)

        return self

    def remove_terminal(self, terminal: Terminal) -> ConductingEquipment:
        """
        Disassociate `terminal` from this `ConductingEquipment`

        `terminal` the `zepben.evolve.cim.iec61970.base.core.terminal.Terminal` to disassociate from this `ConductingEquipment`.
        Returns A reference to this `ConductingEquipment` to allow fluent use.
        Raises `ValueError` if `terminal` was not associated with this `ConductingEquipment`.
        """
        self._terminals.remove(terminal)
        return self

    def clear_terminals(self) -> ConductingEquipment:
        """
        Clear all terminals.
        Returns A reference to this `ConductingEquipment` to allow fluent use.
        """
        self._terminals.clear()
        return self

    def __repr__(self):
        return (f"{super(ConductingEquipment, self).__repr__()}, in_service={self.in_service}, "
                f"normally_in_service={self.normally_in_service}, location={self.location}"
                )

    def _validate_terminal(self, terminal: Terminal) -> bool:
        """
        Validate a terminal against this `ConductingEquipment`'s `zepben.evolve.iec61970.base.core.terminal.Terminal`s.

        `terminal` The `zepben.evolve.iec61970.base.core.terminal.Terminal` to validate.
        Returns True if `zepben.evolve.iec61970.base.core.terminal.Terminal`` is already associated with this `ConductingEquipment`, otherwise False.
        Raises `ValueError` if `zepben.evolve.iec61970.base.core.terminal.Terminal`s `conducting_equipment` is not this `ConductingEquipment`,
        or if this `ConductingEquipment` has a different `zepben.evolve.iec61970.base.core.terminal.Terminal` with the same mRID.
        """
        if self._validate_reference(terminal, self.get_terminal_by_mrid, "A Terminal"):
            return True

        if self._validate_reference_by_sn(terminal.sequence_number, terminal, self.get_terminal_by_sn, "A Terminal"):
            return True

        require(terminal.conducting_equipment is self,
                lambda: f"Terminal {terminal} references another piece of conducting equipment {terminal.conducting_equipment}, expected {str(self)}.")
        return False






#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from enum import Enum

__all__ = ["DiagramStyle"]


class DiagramStyle(Enum):
    """
    The diagram style refer to a style used by the originating system for a diagram.  A diagram style describes
    information such as schematic, geographic, bus-branch etc.
    """

    SCHEMATIC = 0
    """The diagram should be styled as a schematic view."""

    GEOGRAPHIC = 1
    """The diagram should be styled as a geographic view."""

    @property
    def short_name(self):
        return str(self)[13:]



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.evolve.model.cim.iec61970.base.core.conducting_equipment import ConductingEquipment

__all__ = ["EnergyConnection", "RegulatingCondEq"]


class EnergyConnection(ConductingEquipment):
    """
    A connection of energy generation or consumption on the power system phases.
    """
    pass


class RegulatingCondEq(EnergyConnection):
    """
    A short section of conductor with negligible impedance which can be manually removed and replaced if the circuit is
    de-energized. Note that zero-impedance branches can potentially be modeled by other equipment types.
    """

    control_enabled: bool = True
    """Specifies the regulation status of the equipment.  True is regulating, false is not regulating."""



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from enum import Enum

__all__ = ["OrientationKind"]


class OrientationKind(Enum):
    """
    The orientation of the coordinate system with respect to top, left, and the coordinate number system.
    """

    POSITIVE = 0
    """For 2D diagrams, a positive orientation will result in X values increasing from left to right and Y values increasing from bottom to top.  
    This is also known as a right hand orientation."""

    NEGATIVE = 1
    """For 2D diagrams, a negative orientation gives the left-hand orientation (favoured by computer graphics displays) with X values increasing from left to 
    right and Y values increasing from top to bottom. This is also known as a left hand orientation."""

    @property
    def short_name(self):
        return str(self)[16:]



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from enum import Enum
from typing import Optional

from zepben.evolve.model.cim.iec61968.assets.asset import Asset

__all__ = ["Streetlight", "StreetlightLampKind"]


class StreetlightLampKind(Enum):
    """
    Kind of lamp for a `Streetlight`
    """

    UNKNOWN = 0
    HIGH_PRESSURE_SODIUM = 1
    MERCURY_VAPOR = 2
    METAL_HALIDE = 3
    OTHER = 4

    @property
    def short_name(self):
        return str(self)[20:]


class Streetlight(Asset):
    """
    A Streetlight asset.
    """

    pole: Optional[Pole] = None
    """The `zepben.evolve.cim.iec61968.assets.pole.Pole` this Streetlight is attached to."""

    light_rating: int = 0
    """The power rating of the light in watts."""

    lamp_kind: StreetlightLampKind = StreetlightLampKind.UNKNOWN
    """The kind of lamp."""



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from enum import Enum

__all__ = ["SinglePhaseKind", "phasekind_by_id", "SINGLE_PHASE_KIND_VALUES"]


def phasekind_by_id(spk_id):
    """
    Get a SinglePhaseKind by its ID

    `spk_id` ID of the SinglePhaseKind from 0 as per the order of definition
    Returns The SinglePhaseKind
    """
    return SINGLE_PHASE_KIND_VALUES[spk_id]


class SinglePhaseKind(Enum):
    """Enumeration of single phase identifiers. Allows designation of single phases for both transmission and distribution equipment, circuits and loads."""

    NONE = (0, -1)
    """No phase specified"""

    A = (1, 0)
    """Phase A"""

    B = (2, 1)
    """Phase B"""

    C = (3, 2)
    """Phase C"""

    N = (4, 3)
    """Neutral"""

    X = (5, 0)
    """An unknown primary phase."""

    Y = (6, 1)
    """An unknown primary phase."""

    INVALID = (7, -1)
    """Invalid phase. Caused by trying to energise with multiple phases simultaneously."""

    @property
    def bit_mask(self):
        return 1 << self.mask_index if self.mask_index >= 0 else 0

    @property
    def value(self):
        return self.value[0]

    @property
    def mask_index(self):
        return self.value[1]

    @property
    def short_name(self):
        return str(self)[16:]


SINGLE_PHASE_KIND_VALUES = list(SinglePhaseKind.__members__.values())



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from zepben.evolve.model.cim.iec61968.assetinfo.wire_material_kind import WireMaterialKind
from zepben.evolve.model.cim.iec61968.assets.asset_info import AssetInfo

__all__ = ["WireInfo", "CableInfo", "OverheadWireInfo"]


class WireInfo(AssetInfo):
    """
    Wire data that can be specified per line segment phase, or for the line segment as a whole in case its phases all
    have the same wire characteristics

    Attributes -
        rated_current : Current carrying capacity of the wire under stated thermal conditions in amperes.
        material : `zepben.protobuf.cim.iec61968.assetinfo.WireMaterialKind` - Conductor material.
    """
    rated_current: int = 0
    material: WireMaterialKind = WireMaterialKind.UNKNOWN


class CableInfo(WireInfo):
    """
    Cable data. A cable is an underground conductor.
    """
    pass


class OverheadWireInfo(WireInfo):
    """
    Overhead wire data. A "wire" is an above ground conductor.
    """
    pass






#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from enum import Enum

__all__ = ["CustomerKind"]


class CustomerKind(Enum):
    UNKNOWN = 0
    """ Default """

    commercialIndustrial = 1
    """ Commercial industrial customer. """

    energyServiceScheduler = 2
    """ Customer as energy service scheduler. """

    energyServiceSupplier = 3
    """ Customer as energy service supplier. """

    enterprise = 4
    """ --- Missing form CIM --- """

    internalUse = 5
    """ Internal use customer. """

    other = 6
    """ Other kind of customer. """

    pumpingLoad = 7
    """ Pumping load customer. """

    regionalOperator = 8
    """ --- Missing form CIM --- """

    residential = 9
    """ Residential customer. """

    residentialAndCommercial = 10
    """ Residential and commercial customer. """

    residentialAndStreetlight = 11
    """ Residential and streetlight customer. """

    residentialFarmService = 12
    """ Residential farm service customer. """

    residentialStreetlightOthers = 13
    """ Residential streetlight or other related customer. """

    subsidiary = 14
    """ --- Missing form CIM --- """

    windMachine = 15
    """ Wind machine customer. """

    @property
    def short_name(self):
        return str(self)[16:]



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.protobuf.cim.iec61968.common.Agreement_pb2 import Agreement
from zepben.protobuf.cim.iec61968.customers.CustomerAgreement_pb2 import CustomerAgreement
from zepben.protobuf.cim.iec61968.customers.Customer_pb2 import Customer
from zepben.protobuf.cim.iec61968.customers.PricingStructure_pb2 import PricingStructure
from zepben.evolve.services.customer.translator.customer_cim2proto import *
from zepben.evolve.services.customer.translator.customer_proto2cim import *

__all__ = []

Customer.mrid = lambda self: getattr(self, 'or').mrid()
CustomerAgreement.mrid = lambda self: self.agr.mrid()
Agreement.mrid = lambda self: self.doc.mrid()
PricingStructure.mrid = lambda self: self.doc.mrid()

CustomerAgreement.name_and_mrid = lambda self: self.agr.name_and_mrid()
Agreement.name_and_mrid = lambda self: self.doc.name_and_mrid()



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from datetime import datetime
from typing import Optional

from dataclassy import dataclass

__all__ = ["MeasurementValue", "AccumulatorValue", "AnalogValue", "DiscreteValue"]




@dataclass(slots=True)
class MeasurementValue(object):
    """
    The current state for a measurement. A state value is an instance of a measurement from a specific source.
    Measurements can be associated with many state values, each representing a different source for the measurement.
    """
    time_stamp: Optional[datetime] = None
    """The time when the value was last updated."""


class AccumulatorValue(MeasurementValue):
    """AccumulatorValue represents an accumulated (counted) MeasurementValue."""

    value: int = 0
    """The value to supervise"""

    accumulator_mrid: Optional[str] = None
    """The `zepben.evolve.cim.iec61970.base.meas.measurement.Accumulator` mRID of this `AccumulatorValue`"""


class AnalogValue(MeasurementValue):
    """`AnalogValue` represents an analog `MeasurementValue`."""

    value: float = 0.0
    """The value to supervise"""

    analog_mrid: Optional[str] = None
    """The `zepben.evolve.cim.iec61970.base.meas.measurement.Analog` mRID of this `AnalogValue`"""


class DiscreteValue(MeasurementValue):
    """`DiscreteValue` represents a discrete `MeasurementValue`."""

    value: int = 0
    """The value to supervise"""

    discrete_mrid: Optional[str] = None
    """The `zepben.evolve.cim.iec61970.base.meas.measurement.Discrete` mRID of this `DiscreteValue`"""



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

import zepben.evolve.services.common.resolver as resolver

from zepben.evolve.services.diagram.diagrams import DiagramService
from zepben.protobuf.cim.iec61970.base.diagramlayout.Diagram_pb2 import Diagram as PBDiagram
from zepben.protobuf.cim.iec61970.base.diagramlayout.OrientationKind_pb2 import OrientationKind as PBOrientationKind
from zepben.protobuf.cim.iec61970.base.diagramlayout.DiagramStyle_pb2 import DiagramStyle as PBDiagramStyle
from zepben.protobuf.cim.iec61970.base.diagramlayout.DiagramObject_pb2 import DiagramObject as PBDiagramObject
from zepben.protobuf.cim.iec61970.base.diagramlayout.DiagramObjectPoint_pb2 import DiagramObjectPoint as PBDiagramObjectPoint
from zepben.protobuf.cim.iec61970.base.diagramlayout.DiagramObjectStyle_pb2 import DiagramObjectStyle as PBDiagramObjectStyle

from zepben.evolve.model import Diagram, DiagramObject, DiagramObjectPoint
from zepben.evolve.model import DiagramObjectStyle
from zepben.evolve.model import DiagramStyle
from zepben.evolve.model import OrientationKind

__all__ = ["diagramobjectpoint_to_cim", "diagram_to_cim", "diagramobject_to_cim"]


def diagramobjectpoint_to_cim(pb: PBDiagramObjectPoint) -> DiagramObjectPoint:
    return DiagramObjectPoint(pb.xPosition, pb.yPosition)


# IEC61970 DIAGRAM LAYOUT #
def diagram_to_cim(pb: PBDiagram, service: DiagramService):
    cim = Diagram(mrid=pb.mrid(),
                  orientation_kind=OrientationKind[PBOrientationKind.Name(pb.orientationKind)],
                  diagram_style=DiagramStyle[PBDiagramStyle.Name(pb.diagramStyle)])
    for mrid in pb.diagramObjectMRIDs:
        service.resolve_or_defer_reference(resolver.diagram_objects(cim), mrid)
    identifiedobject_to_cim(pb.io, cim, service)
    service.add(cim)


def diagramobject_to_cim(pb: PBDiagramObject, service: DiagramService):
    cim = DiagramObject(mrid=pb.mrid(), identified_object_mrid=pb.identifiedObjectMRID,
                        style=DiagramObjectStyle[PBDiagramObjectStyle.Name(pb.diagramObjectStyle)])
    service.resolve_or_defer_reference(resolver.diagram(cim), pb.diagramMRID)
    for point in pb.diagramObjectPoints:
        cim.add_point(diagramobjectpoint_to_cim(point))
    identifiedobject_to_cim(pb.io, cim, service)
    service.add_diagram_object(cim)


PBDiagram.to_cim = diagram_to_cim
PBDiagramObject.to_cim = diagramobject_to_cim
PBDiagramObjectPoint.to_cim = diagramobjectpoint_to_cim



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Optional

from zepben.evolve.model.cim.iec61968.assetinfo.wire_info import CableInfo
from zepben.evolve.model.cim.iec61970.base.core.conducting_equipment import ConductingEquipment
from zepben.evolve.model.cim.iec61970.base.wires.per_length import PerLengthSequenceImpedance

__all__ = ["AcLineSegment", "Conductor"]


class Conductor(ConductingEquipment):
    """
    Combination of conducting material with consistent electrical characteristics, building a single electrical
    system, used to carry current between points in the power system.
    """

    length: float = 0.0
    """Segment length for calculating line section capabilities."""

    @property
    def wire_info(self):
        """The `zepben.evolve.cim.iec61968.assetinfo.wire_info.WireInfo` for this `Conductor`"""
        return self.asset_info

    def is_underground(self):
        """
        Returns True if this `Conductor` is underground.
        """
        return isinstance(self.wire_info, CableInfo)


class AcLineSegment(Conductor):
    """
    A wire or combination of wires, with consistent electrical characteristics, building a single electrical system,
    used to carry alternating current between points in the power system.

    For symmetrical, transposed 3ph lines, it is sufficient to use attributes of the line segment, which describe
    impedances and admittances for the entire length of the segment. Additionally impedances can be computed by
    using length and associated per length impedances.

    The BaseVoltage at the two ends of ACLineSegments in a Line shall have the same BaseVoltage.nominalVoltage.
    However, boundary lines  may have slightly different BaseVoltage.nominalVoltages and variation is allowed.
    Larger voltage difference in general requires use of an equivalent branch.
    """
    per_length_sequence_impedance: Optional[PerLengthSequenceImpedance] = None
    """A `zepben.evolve.PerLengthSequenceImpedance` describing this ACLineSegment"""




#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from enum import Enum, unique

from zepben.evolve.model.cim.iec61970.base.wires.single_phase_kind import SinglePhaseKind

__all__ = ["PhaseCode", "phasecode_by_id"]


def phasecode_by_id(id: int):
    return _phasecode_members[id]


@unique
class PhaseCode(Enum):
    """
    An unordered enumeration of phase identifiers.  Allows designation of phases for both transmission and distribution equipment,
    circuits and loads. The enumeration, by itself, does not describe how the phases are connected together or connected to ground.
    Ground is not explicitly denoted as a phase.

    Residential and small commercial loads are often served from single-phase, or split-phase, secondary circuits. For example of s12N,
    phases 1 and 2 refer to hot wires that are 180 degrees out of phase, while N refers to the neutral wire. Through single-phase
    transformer connections, these secondary circuits may be served from one or two of the primary phases A, B, and C. For three-phase
    loads, use the A, B, C phase codes instead of s12N.
    """

    NONE = (SinglePhaseKind.NONE,)
    """No phases specified"""

    A = (SinglePhaseKind.A,)
    """Phase A"""

    B = (SinglePhaseKind.B,)
    """Phase B"""

    C = (SinglePhaseKind.C,)
    """Phase C"""

    N = (SinglePhaseKind.N,)
    """Neutral Phase"""

    AB = (SinglePhaseKind.A, SinglePhaseKind.B)
    """Phases A and B"""

    AC = (SinglePhaseKind.A, SinglePhaseKind.C)
    """Phases A and C"""

    AN = (SinglePhaseKind.A, SinglePhaseKind.N)
    """Phases A and N"""

    BC = (SinglePhaseKind.B, SinglePhaseKind.C)
    """Phases B and C"""

    BN = (SinglePhaseKind.B, SinglePhaseKind.N)
    """Phases B and N"""

    CN = (SinglePhaseKind.C, SinglePhaseKind.N)
    """Phases C and N"""

    ABC = (SinglePhaseKind.A, SinglePhaseKind.B, SinglePhaseKind.C)
    """Phases A, B and C"""

    ABN = (SinglePhaseKind.A, SinglePhaseKind.B, SinglePhaseKind.N)
    """Phases A, B and neutral"""

    ACN = (SinglePhaseKind.A, SinglePhaseKind.C, SinglePhaseKind.N)
    """Phases A, C and neutral"""

    BCN = (SinglePhaseKind.B, SinglePhaseKind.C, SinglePhaseKind.N)
    """Phases B, C and neutral"""

    ABCN = (SinglePhaseKind.A, SinglePhaseKind.B, SinglePhaseKind.C, SinglePhaseKind.N)
    """Phases A, B, C and neutral"""

    X = (SinglePhaseKind.X,)
    """Unknown non-neutral phase"""

    XN = (SinglePhaseKind.X, SinglePhaseKind.N)
    """Unknown non-neutral phase plus neutral"""

    XY = (SinglePhaseKind.X, SinglePhaseKind.Y)
    """Two Unknown non-neutral phases"""

    XYN = (SinglePhaseKind.X, SinglePhaseKind.Y, SinglePhaseKind.N)
    """Two Unknown non-neutral phases plus neutral"""

    Y = (SinglePhaseKind.Y,)
    """Unknown non-neutral phase"""

    YN = (SinglePhaseKind.Y, SinglePhaseKind.N)
    """Unknown non-neutral phase plus neutral"""

    @property
    def short_name(self):
        return str(self)[10:]

    @property
    def single_phases(self):
        return self.value[:1]

    @property
    def num_phases(self):
        return len(self.value)


_phasecode_members = list(PhaseCode.__members__.values())



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations
from abc import ABCMeta
from collections import OrderedDict
from dataclassy import dataclass
from typing import Dict, Generator, Callable, Optional, List, Union, Sized

from zepben.evolve.services.common.reference_resolvers import BoundReferenceResolver, UnresolvedReference

__all__ = ["BaseService"]

_GET_DEFAULT = (1,)


@dataclass(slots=True)
class BaseService(object, metaclass=ABCMeta):
    name: str
    _objectsByType: Dict[type, Dict[str, IdentifiedObject]] = OrderedDict()
    _unresolved_references: Dict[str, List[UnresolvedReference]] = OrderedDict()

    def __contains__(self, mrid: str) -> bool:
        """
        Check if `mrid` has any associated object.

        `mrid` The mRID to search for.
        Returns True if there is an object associated with the specified `mrid`, False otherwise.
        """
        for type_map in self._objectsByType.values():
            if mrid in type_map:
                return True
        return False

    def __str__(self):
        return f"{type.__name__}{f' {self.name}' if self.name else ''}"

    def has_unresolved_references(self):
        """
        Returns True if this service has unresolved references, False otherwise.
        """
        return len(self._unresolved_references) > 0

    def len_of(self, t: type = None) -> int:
        """
        Get the len of objects of type `t` in the service.
        `t` The type of object to get the len of. If None (default), will get the len of all objects in the service.
        """
        if t is None:
            return sum([len(vals) for vals in self._objectsByType.values()])
        else:
            return len(self._objectsByType[t].values())

    def num_unresolved_references(self):
        """
        Get the total number of unresolved references.
        Note that this is not terribly cheap call, and should be used sparingly. To test if unresolved references exist,
        use `has_unresolved_references()` instead.
        Returns The number of references in the network that have not already been resolved.
        """
        return len({r.to_mrid for reflist in self._unresolved_references.values() for r in reflist})

    def unresolved_references(self):
        for from_mrid, unresolved_refs in self._unresolved_references.copy().items():
            yield from_mrid, unresolved_refs

    def unresolved_mrids(self):
        seen = set()
        for refs in self._unresolved_references.copy().values():
            for ref in refs:
                if ref.to_mrid not in seen:
                    seen.add(ref.to_mrid)
                    yield ref.to_mrid

    def get(self, mrid: str, type_: type = None, default=_GET_DEFAULT,
            generate_error: Callable[[str, str], str] = lambda mrid, typ: f"Failed to find {typ}[{mrid}]") -> IdentifiedObject:
        """
        Get an object associated with this service.

        `mrid` The mRID of the `iec61970.base.core.identified_object.IdentifiedObject` to retrieve.
        `type_` The `iec61970.base.core.identified_object.IdentifiedObject` subclass type of the object
                      with `mrid`. If None, will check all types stored in the service.
        `default` The default to return if `mrid` can't be found in the service.
        `generate_error` Function to call for an error message. Will be passed the mrid and _type (if set).
        Returns The `iec61970.base.core.identified_object.IdentifiedObject` associated with `mrid`, or default
                 if it is set.
        Raises `KeyError` if `mrid` was not found in the service with `_type` or if no objects of `_type` are
                 stored by the service and default was not set.
        """
        if not mrid:
            raise KeyError("You must specify an mRID to get. Empty/None is invalid.")

        if type_:
            try:
                return self._objectsByType[type_][mrid]
            except KeyError:
                for c, obj_map in self._objectsByType.items():
                    if issubclass(c, type_):
                        try:
                            return obj_map[mrid]
                        except KeyError:
                            pass
                if default is _GET_DEFAULT:
                    raise KeyError(generate_error(mrid, type_.__name__))
                else:
                    return default
        else:
            for object_map in self._objectsByType.values():
                if mrid in object_map:
                    return object_map[mrid]

            if default is _GET_DEFAULT:
                raise KeyError(generate_error(mrid, ""))
            return default

    def __getitem__(self, mrid):
        """
        Get an object associated with this service.
        Note that you should use `get` directly where the type of the desired object is known.
        `mrid` The mRID of the `iec61970.base.core.identified_object.IdentifiedObject` to retrieve.
        Returns The `iec61970.base.core.identified_object.IdentifiedObject` associated with `mrid`.
        Raises `KeyError` if `mrid` was not found in the service with `type`.
        """
        return self.get(mrid)

    def add(self, identified_object: IdentifiedObject) -> bool:
        """
        Associate an object with this service.
        `identified_object` The object to associate with this service.
        Returns True if the object is associated with this service, False otherwise.
        """
        if not identified_object.mrid:
            return False
        # TODO: Only allow supported types

        objs = self._objectsByType.get(identified_object.__class__, dict())
        if identified_object.mrid in objs:
            return False

        # Check other types and make sure this mRID is unique
        for obj_map in self._objectsByType.values():
            if identified_object.mrid in obj_map:
                return False

        unresolved_refs = self._unresolved_references.get(identified_object.mrid, None)
        if unresolved_refs:
            for ref in unresolved_refs:
                ref.resolver.resolve(ref.from_ref, identified_object)
            del self._unresolved_references[identified_object.mrid]

        objs[identified_object.mrid] = identified_object
        self._objectsByType[identified_object.__class__] = objs
        return True

    def resolve_or_defer_reference(self, bound_resolver: BoundReferenceResolver, to_mrid: str) -> bool:
        """
        Resolves a property reference between two types by looking up the `to_mrid` in the service and
        using the provided `bound_resolver` to resolve the reference relationships (including any reverse relationship).

        If the `to_mrid` object has not yet been added to the service, the reference resolution will be deferred until the
        object with `to_mrid` is added to the service, which will then use the resolver from the `bound_resolver` at that
        time to resolve the reference relationship.


        `bound_resolver`
        `to_mrid` The MRID of an object that is the subclass of the to_class of `bound_resolver`.
        Returns true if the reference was resolved, otherwise false if it has been deferred.
        """
        if not to_mrid:
            return True

        from_ = bound_resolver.from_obj
        resolver = bound_resolver.resolver
        reverse_resolver = bound_resolver.reverse_resolver
        try:
            to = self.get(to_mrid, resolver.to_class)
            resolver.resolve(from_, to)
            if reverse_resolver:
                reverse_resolver.resolve(to, from_)

                # Clean up any reverse unresolved references now that the reference has been resolved
                if from_.mrid in self._unresolved_references:
                    refs = self._unresolved_references[from_.mrid]
                    self._unresolved_references[from_.mrid] = [ref for ref in refs if not ref.to_mrid == from_.mrid or not ref.resolver == reverse_resolver]

                    if not self._unresolved_references[from_.mrid]:
                        del self._unresolved_references[from_.mrid]
            return True
        except KeyError:
            urefs = self._unresolved_references.get(to_mrid, list())
            urefs.append(UnresolvedReference(from_ref=from_, to_mrid=to_mrid, resolver=resolver))
            self._unresolved_references[to_mrid] = urefs
            return False

    def get_unresolved_reference_mrids(self, bound_resolvers: Union[BoundReferenceResolver, Sized[BoundReferenceResolver]]) -> Generator[str, None, None]:
        """
        Gets a set of MRIDs that are referenced by the from_obj held by `bound_resolver` that are unresolved.
        `bound_resolver` The `BoundReferenceResolver` to retrieve unresolved references for.
        Returns Set of mRIDs that have unresolved references.
        """
        seen = set()
        try:
            len(bound_resolvers)
            resolvers = bound_resolvers
        except TypeError:
            resolvers = [bound_resolvers]

        for refs in self._unresolved_references.values():
            for ref in refs:
                for resolver in resolvers:
                    if ref.from_ref is resolver.from_obj and ref.to_mrid not in seen and ref.resolver == resolver.resolver:
                        seen.add(ref.to_mrid)
                        yield ref.to_mrid

    def remove(self, identified_object: IdentifiedObject) -> bool:
        """
        Disassociate an object from this service.

        `identified_object` THe object to disassociate from the service.
        Raises `KeyError` if `identified_object` or its type was not present in the service.
        """
        del self._objectsByType[identified_object.__class__][identified_object.mrid]
        return True

    def objects(self, obj_type: Optional[type] = None, exc_types: Optional[List[type]] = None) -> Generator[ IdentifiedObject, None, None]:
        """
        Generator for the objects in this service of type `obj_type`.
        `obj_type` The type of object to yield. If this is a base class it will yield all subclasses.
        Returns Generator over
        """
        if obj_type is None:
            for typ, obj_map in self._objectsByType.items():
                if exc_types:
                    if typ in exc_types:
                        continue
                for obj in obj_map.values():
                    yield obj
            return
        else:
            try:
                for obj in self._objectsByType[obj_type].values():
                    yield obj
            except KeyError:
                for _type, object_map in self._objectsByType.items():
                    if issubclass(_type, obj_type):
                        for obj in object_map.values():
                            yield obj



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional

from zepben.evolve.model.cim.iec61970.base.core.equipment import Equipment

__all__ = ["AuxiliaryEquipment", "FaultIndicator"]


class AuxiliaryEquipment(Equipment):
    """
    `AuxiliaryEquipment` describe equipment that is not performing any primary functions but support for the
    equipment performing the primary function.

    `AuxiliaryEquipment` is attached to primary equipment via an association with `zepben.evolve.cim.iec61970.base.core.Terminal`.
    """
    terminal: Optional[Terminal] = None
    """The `zepben.evolve.iec61970.base.core.terminal.Terminal`` at the `Equipment` where the `AuxiliaryEquipment` is attached."""


class FaultIndicator(AuxiliaryEquipment):
    """
    A FaultIndicator is typically only an indicator (which may or may not be remotely monitored), and not a piece of
    equipment that actually initiates a protection event. It is used for FLISR (Fault Location, Isolation and
    Restoration) purposes, assisting with the dispatch of crews to "most likely" part of the network (i.e. assists
    with determining circuit section where the fault most likely happened).
    """
    pass



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from abc import abstractmethod
from asyncio import get_event_loop
from typing import TypeVar, Union, List, Dict, Type, Callable

from grpc import Channel
from zepben.evolve import NetworkService, BaseService, DiagramService, CustomerService
from zepben.evolve.services.network.translator.network_cim2proto import CimTranslationException
from zepben.evolve.streaming.exceptions import UnsupportedOperationException
from zepben.evolve.streaming.grpc import GrpcClient, GrpcResult
from zepben.evolve.streaming.put.network_rpc import network_rpc_map, diagram_rpc_map, customer_rpc_map
from zepben.protobuf.cp.cp_pb2_grpc import CustomerProducerStub
from zepben.protobuf.cp.cp_requests_pb2 import CreateCustomerServiceRequest, CompleteCustomerServiceRequest
from zepben.protobuf.dp.dp_pb2_grpc import DiagramProducerStub
from zepben.protobuf.dp.dp_requests_pb2 import CompleteDiagramServiceRequest, CreateDiagramServiceRequest
from zepben.protobuf.np.np_pb2_grpc import NetworkProducerStub
from zepben.protobuf.np.np_requests_pb2 import CreateNetworkRequest, CompleteNetworkRequest

__all__ = ["CimProducerClient", "CustomerProducerClient", "NetworkProducerClient", "DiagramProducerClient", "ProducerClient", "SyncProducerClient",
           "SyncCustomerProducerClient", "SyncNetworkProducerClient", "SyncDiagramProducerClient"]
T = TypeVar("T", bound=BaseService)


async def _send(stub, service, rpc_map):
    if not service:
        return

    for obj in service.objects():
        try:
            pb = obj.to_pb()
        except Exception as e:
            raise CimTranslationException(f"Failed to translate {obj} to protobuf.") from e

        try:
            rpc = getattr(stub, rpc_map[type(pb)][0])
        except AttributeError as e:
            raise UnsupportedOperationException(f"RPC {rpc_map[type(pb)][0]} could not be found in {stub.__class__.__name__}") from e

        attrname = f"{obj.__class__.__name__[:1].lower()}{obj.__class__.__name__[1:]}"
        try:
            req = rpc_map[type(pb)][1]()
            getattr(req, attrname).CopyFrom(pb)
        except AttributeError as e:
            raise AttributeError(f"Protobuf attribute {attrname} could not be found - is evolve-grpc in sync?") from e

        rpc(req)


class CimProducerClient(GrpcClient):
    """Base class that defines some helpful functions when producer clients are sending to the server."""


    @abstractmethod
    def send(self, service: T = None):
        """
        Sends objects within the given `service` to the producer server.
                                                                                                                 
        Exceptions that occur during sending will be caught and passed to all error handlers that have been registered. If none of the registered error handlers
        return true to indicate the error has been handled, the exception will be rethrown.
        """
        raise NotImplementedError()


class NetworkProducerClient(CimProducerClient):
    _stub: NetworkProducerStub = None

    def __init__(self, channel=None, stub: NetworkProducerStub = None, error_handlers: List[Callable[[Exception], bool]] = None):
        super().__init__(error_handlers=error_handlers)
        if stub is not None:
            self._stub = stub
        elif channel is not None:
            self._stub = NetworkProducerStub(channel)
        else:
            raise ValueError("Must provide either a channel or a stub")

    async def send(self, service: NetworkService = None):
        """
        Sends objects within the given `service` to the producer server.

        Exceptions that occur during sending will be caught and passed to all error handlers that have been registered. If none of the registered error handlers
        return true to indicate the error has been handled, the exception will be rethrown.
        """
        await self.try_rpc(lambda: self._stub.CreateNetwork(CreateNetworkRequest()))

        await _send(self._stub, service, network_rpc_map)

        await self.try_rpc(lambda: self._stub.CompleteNetwork(CompleteNetworkRequest()))


class DiagramProducerClient(CimProducerClient):
    _stub: DiagramProducerStub = None

    def __init__(self, channel=None, stub: DiagramProducerStub = None, error_handlers: List[Callable[[Exception], bool]] = None):
        super().__init__(error_handlers=error_handlers)
        if stub is not None:
            self._stub = stub
        elif channel is not None:
            self._stub = DiagramProducerStub(channel)
        else:
            raise ValueError("Must provide either a channel or a stub")

    async def send(self, service: DiagramService = None):
        """
        Sends objects within the given `service` to the producer server.

        Exceptions that occur during sending will be caught and passed to all error handlers that have been registered. If none of the registered error handlers
        return true to indicate the error has been handled, the exception will be rethrown.
        """
        await self.try_rpc(lambda: self._stub.CreateDiagramService(CreateDiagramServiceRequest()))

        await _send(self._stub, service, diagram_rpc_map)

        await self.try_rpc(lambda: self._stub.CompleteDiagramService(CompleteDiagramServiceRequest()))


class CustomerProducerClient(CimProducerClient):
    _stub: CustomerProducerStub = None

    def __init__(self, channel=None, stub: CustomerProducerStub = None, error_handlers: List[Callable[[Exception], bool]] = None):
        super().__init__(error_handlers=error_handlers)
        if stub is not None:
            self._stub = stub
        elif channel is not None:
            self._stub = CustomerProducerStub(channel)
        else:
            raise ValueError("Must provide either a channel or a stub")

    async def send(self, service: CustomerService = None):
        """
        Sends objects within the given `service` to the producer server.

        Exceptions that occur during sending will be caught and passed to all error handlers that have been registered. If none of the registered error handlers
        return true to indicate the error has been handled, the exception will be rethrown.
        """
        await self.try_rpc(lambda: self._stub.CreateCustomerService(CreateCustomerServiceRequest()))

        await _send(self._stub, service, customer_rpc_map)

        await self.try_rpc(lambda: self._stub.CompleteCustomerService(CompleteCustomerServiceRequest()))


class ProducerClient(CimProducerClient):
    _channel: Channel = None
    _clients: Dict[Type[BaseService], CimProducerClient] = None

    def __init__(self, channel: Channel = None, clients: Dict[Type[BaseService], CimProducerClient] = None, error_handlers: List[Callable[[Exception], bool]] = None):
        super().__init__(error_handlers=error_handlers)

        if clients is not None:
            self._clients = clients.copy()
        elif channel is not None:
            self._channel = channel
            self._clients = {
                NetworkService: NetworkProducerClient(self._channel),
                DiagramService: DiagramProducerClient(self._channel),
                CustomerService: CustomerProducerClient(self._channel)
            }
        else:
            raise ValueError("You must provide either a channel or clients")

    async def send(self, services: Union[List[BaseService], BaseService] = None):
        """
        Send each service in `services` to the server.

        Exceptions that occur during sending will be caught and passed to all error handlers that have been registered. If none of the registered error handlers
        return true to indicate the error has been handled, the exception will be rethrown.
        """
        if not services:
            return GrpcResult(UnsupportedOperationException("No services were provided"))

        sent = []
        for service in services:
            client = self._clients[type(service)]
            await client.send(service)
            sent.append(type(service))

        for s in self._clients.keys():
            if s not in sent:
                client = self._clients[s]
                await client.send()


class SyncProducerClient(ProducerClient):

    def send(self, services: Union[List[BaseService], BaseService] = None):
        return get_event_loop().run_until_complete(super().send(services))


class SyncCustomerProducerClient(CustomerProducerClient):

    def send(self, service: CustomerService = None):
        return get_event_loop().run_until_complete(super().send(service))


class SyncNetworkProducerClient(NetworkProducerClient):

    def send(self, service: NetworkService = None):
        return get_event_loop().run_until_complete(super().send(service))


class SyncDiagramProducerClient(DiagramProducerClient):

    def send(self, service: DiagramService = None):
        return get_event_loop().run_until_complete(super().send(service))



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import List, Optional, Generator

from zepben.evolve.model.cim.iec61970.base.wires.energy_connection import EnergyConnection
from zepben.evolve.model.cim.iec61970.base.wires.energy_source_phase import EnergySourcePhase
from zepben.evolve.util import nlen, get_by_mrid, ngen, safe_remove

__all__ = ["EnergySource"]


class EnergySource(EnergyConnection):
    """
    A generic equivalent for an energy supplier on a transmission or distribution voltage level.
    """

    _energy_source_phases: Optional[List[EnergySourcePhase]] = None
    active_power: float = 0.0
    """High voltage source active injection. Load sign convention is used, i.e. positive sign means flow out from a node. Starting value for steady state solutions"""

    r: float = 0.0
    """Positive sequence Thevenin resistance."""

    x: float = 0.0
    """Positive sequence Thevenin reactance."""

    reactive_power: float = 0.0
    """High voltage source reactive injection. Load sign convention is used, i.e. positive sign means flow out from a node. 
    Starting value for steady state solutions."""

    voltage_angle: float = 0.0
    """Phase angle of a-phase open circuit."""

    voltage_magnitude: float = 0.0
    """Phase-to-phase open circuit voltage magnitude."""

    p_max: float = 0.0
    p_min: float = 0.0
    r0: float = 0.0
    rn: float = 0.0
    x0: float = 0.0
    xn: float = 0.0

    def __init__(self, usage_points: List[UsagePoint] = None, equipment_containers: List[EquipmentContainer] = None,
                 operational_restrictions: List[OperationalRestriction] = None, current_feeders: List[Feeder] = None, terminals: List[Terminal] = None,
                 energy_source_phases: List[EnergySourcePhase] = None):
        super(EnergySource, self).__init__(usage_points=usage_points, equipment_containers=equipment_containers,
                                           operational_restrictions=operational_restrictions,
                                           current_feeders=current_feeders, terminals=terminals)
        if energy_source_phases:
            for phase in energy_source_phases:
                self.add_phase(phase)

    @property
    def phases(self) -> Generator[EnergySourcePhase, None, None]:
        """
        The `EnergySourcePhase`s for this `EnergySource`.
        """
        return ngen(self._energy_source_phases)

    def has_phases(self):
        """
        Check if this source has any associated `EnergySourcePhase`s
        Returns True if there is at least one `EnergySourcePhase`, otherwise False
        """
        return nlen(self._energy_source_phases) > 0

    def num_phases(self):
        """Return the number of `EnergySourcePhase`s associated with this `EnergySource`"""
        return nlen(self._energy_source_phases)

    def get_phase(self, mrid: str) -> EnergySource:
        """
        Get the `zepben.evolve.cim.iec61970.base.wires.energy_source_phase.EnergySourcePhase` for this `EnergySource` identified by `mrid`

        `mrid` the mRID of the required `zepben.evolve.cim.iec61970.base.wires.energy_source_phase.EnergySourcePhase`
        Returns The `zepben.evolve.cim.iec61970.base.wires.energy_source_phase.EnergySourcePhase` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._energy_source_phases, mrid)

    def add_phase(self, phase: EnergySourcePhase) -> EnergySource:
        """
        Associate an `zepben.evolve.cim.iec61970.base.wires.energy_source_phase.EnergySourcePhase` with this `EnergySource`

        `phase` the `EnergySourcePhase` to associate with this `EnergySource`.
        Returns A reference to this `EnergySource` to allow fluent use.
        Raises `ValueError` if another `EnergySourcePhase` with the same `mrid` already exists for this `EnergySource`.
        """
        if self._validate_reference(phase, self.get_phase, "An EnergySourcePhase"):
            return self
        self._energy_source_phases = list() if self._energy_source_phases is None else self._energy_source_phases
        self._energy_source_phases.append(phase)
        return self

    def remove_phases(self, phase: EnergySourcePhase) -> EnergySource:
        """
        Disassociate an `phase` from this `EnergySource`

        `phase` the `EnergySourcePhase` to disassociate from this `EnergySource`.
        Returns A reference to this `EnergySource` to allow fluent use.
        Raises `ValueError` if `phase` was not associated with this `EnergySource`.
        """
        self._energy_source_phases = safe_remove(self._energy_source_phases, phase)
        return self

    def clear_phases(self) -> EnergySource:
        """
        Clear all phases.
        Returns A reference to this `EnergySource` to allow fluent use.
        """
        self._energy_source_phases = None
        return self



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.protobuf.cim.iec61968.assetinfo.CableInfo_pb2 import CableInfo
from zepben.protobuf.cim.iec61968.assetinfo.WireInfo_pb2 import WireInfo
from zepben.protobuf.cim.iec61968.assetinfo.OverheadWireInfo_pb2 import OverheadWireInfo
from zepben.protobuf.cim.iec61968.assets.AssetContainer_pb2 import AssetContainer
from zepben.protobuf.cim.iec61968.assets.AssetInfo_pb2 import AssetInfo
from zepben.protobuf.cim.iec61968.assets.AssetOrganisationRole_pb2 import AssetOrganisationRole
from zepben.protobuf.cim.iec61968.assets.AssetOwner_pb2 import AssetOwner
from zepben.protobuf.cim.iec61968.assets.Asset_pb2 import Asset
from zepben.protobuf.cim.iec61968.assets.Pole_pb2 import Pole
from zepben.protobuf.cim.iec61968.assets.Streetlight_pb2 import Streetlight
from zepben.protobuf.cim.iec61968.assets.Structure_pb2 import Structure
from zepben.protobuf.cim.iec61968.common.Location_pb2 import Location
from zepben.protobuf.cim.iec61968.metering.EndDevice_pb2 import EndDevice
from zepben.protobuf.cim.iec61968.metering.Meter_pb2 import Meter
from zepben.protobuf.cim.iec61968.metering.UsagePoint_pb2 import UsagePoint
from zepben.protobuf.cim.iec61968.operations.OperationalRestriction_pb2 import OperationalRestriction
from zepben.protobuf.cim.iec61970.base.auxiliaryequipment.AuxiliaryEquipment_pb2 import AuxiliaryEquipment
from zepben.protobuf.cim.iec61970.base.auxiliaryequipment.FaultIndicator_pb2 import FaultIndicator
from zepben.protobuf.cim.iec61970.base.core.AcDcTerminal_pb2 import AcDcTerminal
from zepben.protobuf.cim.iec61970.base.core.BaseVoltage_pb2 import BaseVoltage
from zepben.protobuf.cim.iec61970.base.core.ConductingEquipment_pb2 import ConductingEquipment
from zepben.protobuf.cim.iec61970.base.core.ConnectivityNodeContainer_pb2 import ConnectivityNodeContainer
from zepben.protobuf.cim.iec61970.base.core.ConnectivityNode_pb2 import ConnectivityNode
from zepben.protobuf.cim.iec61970.base.core.EquipmentContainer_pb2 import EquipmentContainer
from zepben.protobuf.cim.iec61970.base.core.Equipment_pb2 import Equipment
from zepben.protobuf.cim.iec61970.base.core.Feeder_pb2 import Feeder
from zepben.protobuf.cim.iec61970.base.core.GeographicalRegion_pb2 import GeographicalRegion
from zepben.protobuf.cim.iec61970.base.core.PowerSystemResource_pb2 import PowerSystemResource
from zepben.protobuf.cim.iec61970.base.core.Site_pb2 import Site
from zepben.protobuf.cim.iec61970.base.core.SubGeographicalRegion_pb2 import SubGeographicalRegion
from zepben.protobuf.cim.iec61970.base.core.Substation_pb2 import Substation
from zepben.protobuf.cim.iec61970.base.core.Terminal_pb2 import Terminal
from zepben.protobuf.cim.iec61970.base.meas.Accumulator_pb2 import Accumulator
from zepben.protobuf.cim.iec61970.base.meas.Analog_pb2 import Analog
from zepben.protobuf.cim.iec61970.base.meas.Control_pb2 import Control
from zepben.protobuf.cim.iec61970.base.meas.Discrete_pb2 import Discrete
from zepben.protobuf.cim.iec61970.base.meas.IoPoint_pb2 import IoPoint
from zepben.protobuf.cim.iec61970.base.meas.Measurement_pb2 import Measurement
from zepben.protobuf.cim.iec61970.base.scada.RemoteControl_pb2 import RemoteControl
from zepben.protobuf.cim.iec61970.base.scada.RemotePoint_pb2 import RemotePoint
from zepben.protobuf.cim.iec61970.base.scada.RemoteSource_pb2 import RemoteSource
from zepben.protobuf.cim.iec61970.base.wires.AcLineSegment_pb2 import AcLineSegment
from zepben.protobuf.cim.iec61970.base.wires.Breaker_pb2 import Breaker
from zepben.protobuf.cim.iec61970.base.wires.Conductor_pb2 import Conductor
from zepben.protobuf.cim.iec61970.base.wires.Connector_pb2 import Connector
from zepben.protobuf.cim.iec61970.base.wires.Disconnector_pb2 import Disconnector
from zepben.protobuf.cim.iec61970.base.wires.EnergyConnection_pb2 import EnergyConnection
from zepben.protobuf.cim.iec61970.base.wires.EnergyConsumerPhase_pb2 import EnergyConsumerPhase
from zepben.protobuf.cim.iec61970.base.wires.EnergyConsumer_pb2 import EnergyConsumer
from zepben.protobuf.cim.iec61970.base.wires.EnergySourcePhase_pb2 import EnergySourcePhase
from zepben.protobuf.cim.iec61970.base.wires.EnergySource_pb2 import EnergySource
from zepben.protobuf.cim.iec61970.base.wires.Fuse_pb2 import Fuse
from zepben.protobuf.cim.iec61970.base.wires.Jumper_pb2 import Jumper
from zepben.protobuf.cim.iec61970.base.wires.Junction_pb2 import Junction
from zepben.protobuf.cim.iec61970.base.wires.Line_pb2 import Line
from zepben.protobuf.cim.iec61970.base.wires.LinearShuntCompensator_pb2 import LinearShuntCompensator
from zepben.protobuf.cim.iec61970.base.wires.PerLengthImpedance_pb2 import PerLengthImpedance
from zepben.protobuf.cim.iec61970.base.wires.PerLengthLineParameter_pb2 import PerLengthLineParameter
from zepben.protobuf.cim.iec61970.base.wires.PerLengthSequenceImpedance_pb2 import PerLengthSequenceImpedance
from zepben.protobuf.cim.iec61970.base.wires.PowerTransformerEnd_pb2 import PowerTransformerEnd
from zepben.protobuf.cim.iec61970.base.wires.PowerTransformer_pb2 import PowerTransformer
from zepben.protobuf.cim.iec61970.base.wires.ProtectedSwitch_pb2 import ProtectedSwitch
from zepben.protobuf.cim.iec61970.base.wires.RatioTapChanger_pb2 import RatioTapChanger
from zepben.protobuf.cim.iec61970.base.wires.Recloser_pb2 import Recloser
from zepben.protobuf.cim.iec61970.base.wires.RegulatingCondEq_pb2 import RegulatingCondEq
from zepben.protobuf.cim.iec61970.base.wires.ShuntCompensator_pb2 import ShuntCompensator
from zepben.protobuf.cim.iec61970.base.wires.Switch_pb2 import Switch
from zepben.protobuf.cim.iec61970.base.wires.TapChanger_pb2 import TapChanger
from zepben.protobuf.cim.iec61970.base.wires.TransformerEnd_pb2 import TransformerEnd
from zepben.protobuf.cim.iec61970.infiec61970.feeder.Loop_pb2 import Loop
from zepben.protobuf.cim.iec61970.infiec61970.feeder.Circuit_pb2 import Circuit

from zepben.evolve.services.network.translator.network_proto2cim import *
from zepben.evolve.services.network.translator.network_cim2proto import *
__all__ = []

CableInfo.mrid = lambda self: self.wi.mrid()
OverheadWireInfo.mrid = lambda self: self.wi.mrid()
WireInfo.mrid = lambda self: self.ai.mrid()
Asset.mrid = lambda self: self.io.mRID
AssetContainer.mrid = lambda self: self.at.mrid()
AssetInfo.mrid = lambda self: self.io.mRID
AssetOrganisationRole.mrid = lambda self: getattr(self, "or").mrid()
AssetOwner.mrid = lambda self: self.aor.mrid()
Pole.mrid = lambda self: self.st.mrid()
Streetlight.mrid = lambda self: self.at.mrid()
Structure.mrid = lambda self: self.ac.mrid()
Location.mrid = lambda self: self.io.mRID
EndDevice.mrid = lambda self: self.ac.mrid()
Meter.mrid = lambda self: self.ed.mrid()
UsagePoint.mrid = lambda self: self.io.mRID
OperationalRestriction.mrid = lambda self: self.doc.mrid()
AuxiliaryEquipment.mrid = lambda self: self.eq.mrid()
FaultIndicator.mrid = lambda self: self.ae.mrid()
AcDcTerminal.mrid = lambda self: self.io.mRID
BaseVoltage.mrid = lambda self: self.io.mRID
ConductingEquipment.mrid = lambda self: self.eq.mrid()
ConnectivityNode.mrid = lambda self: self.io.mRID
ConnectivityNodeContainer.mrid = lambda self: self.psr.mrid()
Equipment.mrid = lambda self: self.psr.mrid()
EquipmentContainer.mrid = lambda self: self.cnc.mrid()
Feeder.mrid = lambda self: self.ec.mrid()
GeographicalRegion.mrid = lambda self: self.io.mRID
PowerSystemResource.mrid = lambda self: self.io.mRID
Site.mrid = lambda self: self.ec.mrid()
SubGeographicalRegion.mrid = lambda self: self.io.mRID
Substation.mrid = lambda self: self.ec.mrid()
Terminal.mrid = lambda self: self.ad.mrid()
AcLineSegment.mrid = lambda self: self.cd.mrid()
Breaker.mrid = lambda self: self.sw.mrid()
Conductor.mrid = lambda self: self.ce.mrid()
Connector.mrid = lambda self: self.ce.mrid()
Disconnector.mrid = lambda self: self.sw.mrid()
EnergyConnection.mrid = lambda self: self.ce.mrid()
EnergyConsumer.mrid = lambda self: self.ec.mrid()
EnergyConsumerPhase.mrid = lambda self: self.psr.mrid()
EnergySource.mrid = lambda self: self.ec.mrid()
EnergySourcePhase.mrid = lambda self: self.psr.mrid()
Fuse.mrid = lambda self: self.sw.mrid()
Jumper.mrid = lambda self: self.sw.mrid()
Junction.mrid = lambda self: self.cn.mrid()
Line.mrid = lambda self: self.ec.mrid()
LinearShuntCompensator.mrid = lambda self: self.sc.mrid()
PerLengthImpedance.mrid = lambda self: self.lp.mrid()
PerLengthLineParameter.mrid = lambda self: self.io.mRID
PerLengthSequenceImpedance.mrid = lambda self: self.pli.mrid()
PowerTransformer.mrid = lambda self: self.ce.mrid()
PowerTransformerEnd.mrid = lambda self: self.te.mrid()
ProtectedSwitch.mrid = lambda self: self.sw.mrid()
RatioTapChanger.mrid = lambda self: self.tc.mrid()
Recloser.mrid = lambda self: self.sw.mrid()
RegulatingCondEq.mrid = lambda self: self.ec.mrid()
ShuntCompensator.mrid = lambda self: self.rce.mrid()
Switch.mrid = lambda self: self.ce.mrid()
TapChanger.mrid = lambda self: self.psr.mrid()
TransformerEnd.mrid = lambda self: self.io.mRID
Loop.mrid = lambda self: self.io.mRID
Circuit.mrid = lambda self: self.l.mrid()
Accumulator.mrid = lambda self: self.measurement.mrid()
Analog.mrid = lambda self: self.measurement.mrid()
Discrete.mrid = lambda self: self.measurement.mrid()
Control.mrid = lambda self: self.ip.mrid()
IoPoint.mrid = lambda self: self.io.mRID
Measurement.mrid = lambda self: self.io.mRID
RemoteControl.mrid = lambda self: self.rp.mrid()
RemotePoint.mrid = lambda self: self.io.mRID
RemoteSource.mrid = lambda self: self.rp.mrid()


PowerSystemResource.name_and_mrid = lambda self: self.io.name_and_mrid()
ConductingEquipment.name_and_mrid = lambda self: self.eq.name_and_mrid()
Equipment.name_and_mrid = lambda self: self.psr.name_and_mrid()
ConnectivityNodeContainer.name_and_mrid = lambda self: self.psr.name_and_mrid()
EquipmentContainer.name_and_mrid = lambda self: self.cnc.name_and_mrid()
Feeder.name_and_mrid = lambda self: self.ec.name_and_mrid()
EnergyConsumerPhase.name_and_mrid = lambda self: self.psr.name_and_mrid()
EnergySourcePhase.name_and_mrid = lambda self: self.psr.name_and_mrid()
PowerTransformerEnd.name_and_mrid = lambda self: self.te.name_and_mrid()
AcDcTerminal.name_and_mrid = lambda self: self.io.name_and_mrid()
TransformerEnd.name_and_mrid = lambda self: self.io.name_and_mrid()
Terminal.name_and_mrid = lambda self: self.ad.name_and_mrid()

# location_mrid
PowerSystemResource.location_mrid = lambda self: getattr(self, "locationMRID", None)
Equipment.location_mrid = lambda self: self.psr.location_mrid()
AuxiliaryEquipment.location_mrid = lambda self: self.eq.location_mrid()
FaultIndicator.location_mrid = lambda self: self.ae.location_mrid()
ConductingEquipment.location_mrid = lambda self: self.eq.location_mrid()
Conductor.location_mrid = lambda self: self.ce.location_mrid()
Connector.location_mrid = lambda self: self.ce.location_mrid()
EnergyConnection.location_mrid = lambda self: self.ce.location_mrid()
PowerTransformer.location_mrid = lambda self: self.ce.location_mrid()
Switch.location_mrid = lambda self: self.ce.location_mrid()
AcLineSegment.location_mrid = lambda self: self.cd.location_mrid()
Junction.location_mrid = lambda self: self.cn.location_mrid()
EnergyConsumer.location_mrid = lambda self: self.ec.location_mrid()
EnergySource.location_mrid = lambda self: self.ec.location_mrid()
RegulatingCondEq.location_mrid = lambda self: self.ec.location_mrid()
Disconnector.location_mrid = lambda self: self.sw.location_mrid()
Fuse.location_mrid = lambda self: self.sw.location_mrid()
Jumper.location_mrid = lambda self: self.sw.location_mrid()
ProtectedSwitch.location_mrid = lambda self: self.sw.location_mrid()
ShuntCompensator.location_mrid = lambda self: self.rce.location_mrid()
Breaker.location_mrid = lambda self: self.sw.location_mrid()
Recloser.location_mrid = lambda self: self.sw.location_mrid()
LinearShuntCompensator.location_mrid = lambda self: self.sc.location_mrid()

# service_location_mrid
EndDevice.service_location_mrid = lambda self: getattr(self, "serviceLocationMRID", None)
Meter.service_location_mrid = lambda self: self.ed.service_location_mrid()

# usage_point_location_mrid
UsagePoint.usage_point_location_mrid = lambda self: getattr(self, "usagePointLocationMRID", None)

# terminal_mrid
AuxiliaryEquipment.terminal_mrid = lambda self: getattr(self, "terminalMRID", None)
FaultIndicator.terminal_mrid = lambda self: self.ae.terminal_mrid()

# terminal_mrids
ConductingEquipment.terminal_mrids = lambda self: getattr(self, "terminalMRIDs", [])
Conductor.terminal_mrids = lambda self: self.ce.terminal_mrids()
Connector.terminal_mrids = lambda self: self.ce.terminal_mrids()
EnergyConnection.terminal_mrids = lambda self: self.ce.terminal_mrids()
PowerTransformer.terminal_mrids = lambda self: self.ce.terminal_mrids()
Switch.terminal_mrids = lambda self: self.ce.terminal_mrids()
AcLineSegment.terminal_mrids = lambda self: self.cd.terminal_mrids()
Junction.terminal_mrids = lambda self: self.cn.terminal_mrids()
EnergyConsumer.terminal_mrids = lambda self: self.ec.terminal_mrids()
EnergySource.terminal_mrids = lambda self: self.ec.terminal_mrids()
RegulatingCondEq.terminal_mrids = lambda self: self.ec.terminal_mrids()
Disconnector.terminal_mrids = lambda self: self.sw.terminal_mrids()
Fuse.terminal_mrids = lambda self: self.sw.terminal_mrids()
Jumper.terminal_mrids = lambda self: self.sw.terminal_mrids()
ProtectedSwitch.terminal_mrids = lambda self: self.sw.terminal_mrids()
ShuntCompensator.terminal_mrids = lambda self: self.rce.terminal_mrids()
Breaker.terminal_mrids = lambda self: self.sw.terminal_mrids()
Recloser.terminal_mrids = lambda self: self.sw.terminal_mrids()
LinearShuntCompensator.terminal_mrids = lambda self: self.sc.terminal_mrids()

# base_voltage_mrid
ConductingEquipment.base_voltage_mrid = lambda self: getattr(self, "baseVoltageMRID", None)
Conductor.base_voltage_mrid = lambda self: self.ce.base_voltage_mrid()
Connector.base_voltage_mrid = lambda self: self.ce.base_voltage_mrid()
EnergyConnection.base_voltage_mrid = lambda self: self.ce.base_voltage_mrid()
PowerTransformer.base_voltage_mrid = lambda self: self.ce.base_voltage_mrid()
Switch.base_voltage_mrid = lambda self: self.ce.base_voltage_mrid()
AcLineSegment.base_voltage_mrid = lambda self: self.cd.base_voltage_mrid()
Junction.base_voltage_mrid = lambda self: self.cn.base_voltage_mrid()
EnergyConsumer.base_voltage_mrid = lambda self: self.ec.base_voltage_mrid()
EnergySource.base_voltage_mrid = lambda self: self.ec.base_voltage_mrid()
RegulatingCondEq.base_voltage_mrid = lambda self: self.ec.base_voltage_mrid()
Disconnector.base_voltage_mrid = lambda self: self.sw.base_voltage_mrid()
Fuse.base_voltage_mrid = lambda self: self.sw.base_voltage_mrid()
Jumper.base_voltage_mrid = lambda self: self.sw.base_voltage_mrid()
ProtectedSwitch.base_voltage_mrid = lambda self: self.sw.base_voltage_mrid()
ShuntCompensator.base_voltage_mrid = lambda self: self.rce.base_voltage_mrid()
Breaker.base_voltage_mrid = lambda self: self.sw.base_voltage_mrid()
Recloser.base_voltage_mrid = lambda self: self.sw.base_voltage_mrid()
LinearShuntCompensator.base_voltage_mrid = lambda self: self.sc.base_voltage_mrid()

# normal_energizing_substation_mrid
Feeder.normal_energizing_substation_mrid = lambda self: getattr(self, "normalEnergizingSubstationMRID", None)

# per_length_sequence_impedance_mrid
AcLineSegment.per_length_sequence_impedance_mrid = lambda self: getattr(self, "perLengthSequenceImpedanceMRID", None)

# asset_info_mrid
Conductor.asset_info_mrid = lambda self: getattr(self, "assetInfoMRID", None)
AcLineSegment.asset_info_mrid = lambda self: self.cd.asset_info_mrid()

# ratio_tap_changer_mrid
TransformerEnd.ratio_tap_changer_mrid = lambda self: getattr(self, "ratioTapChangerMRID", None)
PowerTransformerEnd.ratio_tap_changer_mrid = lambda self: self.te.ratio_tap_changer_mrid()


#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.protobuf.cim.iec61970.base.diagramlayout.DiagramObject_pb2 import DiagramObject
from zepben.protobuf.cim.iec61970.base.diagramlayout.Diagram_pb2 import Diagram
from zepben.evolve.services.diagram.translator.diagram_proto2cim import *
from zepben.evolve.services.diagram.translator.diagram_cim2proto import *

__all__ = []


Diagram.mrid = lambda self: self.io.mRID
DiagramObject.mrid = lambda self: self.io.mRID



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations
import re
import os
from collections.abc import Sized
from typing import Set, List, Optional, Iterable, Callable, Any, TypeVar, Generator
from uuid import UUID

T = TypeVar('T')


def snake2camelback(name):
    return ''.join(word.title() for word in name.split('_'))


_camel_pattern = re.compile(r'(?<!^)(?=[A-Z])')


def camel2snake(name):
    return _camel_pattern.sub('_', name).lower()


def iter_but_not_str(obj):
    return isinstance(obj, Iterable) and not isinstance(obj, (str, bytes, bytearray, dict))


def get_by_mrid(collection: Optional[Iterable[IdentifiedObject]], mrid: str) -> IdentifiedObject:
    """
    Get an `zepben.evolve.cim.iec61970.base.core.identified_object.IdentifiedObject` from `collection` based on
    its mRID.
    `collection` The collection to operate on
    `mrid` The mRID of the `IdentifiedObject` to lookup in the collection
    Returns The `IdentifiedObject`
    Raises `KeyError` if `mrid` was not found in the collection.
    """
    if not collection:
        raise KeyError(mrid)
    for io in collection:
        if io.mrid == mrid:
            return io
    raise KeyError(mrid)


def contains_mrid(collection: Optional[Iterable[IdentifiedObject]], mrid: str) -> bool:
    """
    Check if a collection of `zepben.evolve.cim.iec61970.base.core.identified_object.IdentifiedObject` contains an
    object with a specified mRID.
    `collection` The collection to operate on
    `mrid` The mRID to look up.
    Returns True if an `IdentifiedObject` is found in the collection with the specified mRID, False otherwise.
    """
    if not collection:
        return False
    try:
        if get_by_mrid(collection, mrid):
            return True
    except KeyError:
        return False


def safe_remove(collection: Optional[List], obj: IdentifiedObject):
    """
    Remove an IdentifiedObject from a collection safely.
    Raises `ValueError` if `obj` is not in the collection.
    Returns The collection if successfully removed or None if after removal the collection was empty.
    """
    if collection is not None:
        collection.remove(obj)
        if not collection:
            return None
        return collection
    else:
        raise ValueError(obj)


def nlen(sized: Optional[Sized]) -> int:
    """
    Get the len of a nullable sized type.
    `sized` The object to get length of
    Returns 0 if `sized` is None, otherwise len(`sized`)
    """
    return 0 if sized is None else len(sized)


def ngen(collection: Optional[Iterable[T]]) -> Generator[T, None, None]:
    if collection:
        for item in collection:
            yield item


def is_none_or_empty(sized: Optional[Sized]) -> bool:
    """
    Check if a given object is empty and return None if it is.
    `sized` Any type implementing `__len__`
    Returns `sized` if len(sized) > 0, or None if sized is None or len(sized) == 0.
    """
    return sized is None or not len(sized)


def require(condition: bool, lazy_message: Callable[[], Any]):
    """
    Raise a `ValueError` if condition is not met, with the result of calling `lazy_message` as the message,
    if the result is false.
    """
    if not condition:
        raise ValueError(str(lazy_message()))


def pb_or_none(cim: Optional[Any]):
    """ Convert to a protobuf type or return None if cim was None """
    return cim.to_pb() if cim is not None else None


class CopyableUUID(UUID):
    
    def __init__(self):
        super().__init__(bytes=os.urandom(16), version=4)

    def copy(self):
        return UUID(bytes=os.urandom(16), version=4)





#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.evolve.model.cim.iec61970.base.core.identified_object import IdentifiedObject

__all__ = ["RemotePoint"]


class RemotePoint(IdentifiedObject):
    """
    For a RTU remote points correspond to telemetered values or control outputs. Other units (e.g. control centers)
    usually also contain calculated values.
    """
    pass



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from datetime import datetime
from typing import Optional

from zepben.evolve.model.cim.iec61970.base.core.identified_object import IdentifiedObject

__all__ = ["Document", "Agreement"]


class Document(IdentifiedObject):
    """
    Parent class for different groupings of information collected and managed as a part of a business process.
    It will frequently contain references to other objects, such as assets, people and power system resources.
    """
    title: str = ""
    """Document title."""

    created_date_time: Optional[datetime] = None
    """Date and time that this document was created."""

    author_name: str = ""
    """Name of the author of this document."""

    type: str = ""
    """Utility-specific classification of this document, according to its corporate standards, practices, 
    and existing IT systems (e.g., for management of assets, maintenance, work, outage, customers, etc.)."""

    status: str = ""
    """Status of subject matter (e.g., Agreement, Work) this document represents."""

    comment: str = ""
    """Free text comment"""


class Agreement(Document):
    """
    Formal agreement between two parties defining the terms and conditions for a set of services. The specifics of
    the services are, in turn, defined via one or more service agreements.
    """
    pass



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional, Generator, List

from zepben.evolve.model.cim.iec61968.common.document import Document
from zepben.evolve.util import get_by_mrid, nlen, ngen, safe_remove

__all__ = ["OperationalRestriction"]


class OperationalRestriction(Document):
    """
    A document that can be associated with equipment to describe any sort of restrictions compared with the
    original manufacturer's specification or with the usual operational practice e.g.
    temporary maximum loadings, maximum switching current, do not operate if bus couplers are open, etc.

    In the UK, for example, if a breaker or switch ever mal-operates, this is reported centrally and utilities
    use their asset systems to identify all the installed devices of the same manufacturer's type.
    They then apply operational restrictions in the operational systems to warn operators of potential problems.
    After appropriate inspection and maintenance, the operational restrictions may be removed.
    """
    _equipment: Optional[List[Equipment]] = None

    def __init__(self, equipment: List[Equipment] = None):
        if equipment:
            for eq in equipment:
                self.add_equipment(eq)

    @property
    def num_equipment(self):
        """
        Returns the number of `zepben.evolve.cim.iec61970.base.core.equipment.Equipment` associated with this `OperationalRestriction`
        """
        return nlen(self._equipment)

    @property
    def equipment(self) -> Generator[Equipment, None, None]:
        """
        The `zepben.evolve.cim.iec61970.base.core.equipment.Equipment` to which this `OperationalRestriction` applies.
        """
        return ngen(self._equipment)

    def get_equipment(self, mrid: str) -> Equipment:
        """
        Get the `zepben.evolve.cim.iec61970.base.core.equipment.Equipment` for this `OperationalRestriction` identified by `mrid`

        `mrid` The mRID of the required `zepben.evolve.cim.iec61970.base.core.equipment.Equipment`
        Returns The `zepben.evolve.cim.iec61970.base.core.equipment.Equipment` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._equipment, mrid)

    def add_equipment(self, equipment: Equipment) -> OperationalRestriction:
        """
        Associate an `zepben.evolve.cim.iec61970.base.core.equipment.Equipment` with this `OperationalRestriction`

        `equipment` The `zepben.evolve.cim.iec61970.base.core.equipment.Equipment` to associate with this `OperationalRestriction`.
        Returns A reference to this `OperationalRestriction` to allow fluent use.
        Raises `ValueError` if another `Equipment` with the same `mrid` already exists for this `OperationalRestriction`.
        """
        if self._validate_reference(equipment, self.get_equipment, "An Equipment"):
            return self
        self._equipment = list() if self._equipment is None else self._equipment
        self._equipment.append(equipment)
        return self

    def remove_equipment(self, equipment: Equipment) -> OperationalRestriction:
        """
        Disassociate `equipment` from this `OperationalRestriction`.

        `equipment` The `zepben.evolve.cim.iec61970.base.core.equipment.Equipment` to disassociate from this `OperationalRestriction`.
        Returns A reference to this `OperationalRestriction` to allow fluent use.
        Raises `ValueError` if `equipment` was not associated with this `OperationalRestriction`.
        """
        self._equipment = safe_remove(self._equipment, equipment)
        return self

    def clear_equipment(self) -> OperationalRestriction:
        """
        Clear all equipment.
        Returns A reference to this `OperationalRestriction` to allow fluent use.
        """
        self._equipment = None
        return self



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.evolve.model.cim.iec61968.customers.pricing_structure import *
from zepben.evolve.model.cim.iec61968.customers.customer_agreement import *
from zepben.evolve.model.cim.iec61968.customers.customer_kind import *
from zepben.evolve.model.cim.iec61968.customers.customer import *
from zepben.evolve.model.cim.iec61968.customers.tariff import *
from zepben.evolve.model.cim.iec61968.assets.structure import *
from zepben.evolve.model.cim.iec61968.assets.asset import *
from zepben.evolve.model.cim.iec61968.assets.pole import *
from zepben.evolve.model.cim.iec61968.assets.asset_organisation_role import *
from zepben.evolve.model.cim.iec61968.assets.asset_info import *
from zepben.evolve.model.cim.iec61968.assets.streetlight import *
from zepben.evolve.model.cim.iec61968.operations.operational_restriction import *
from zepben.evolve.model.cim.iec61968.assetinfo.wire_info import *
from zepben.evolve.model.cim.iec61968.assetinfo.wire_material_kind import *
from zepben.evolve.model.cim.iec61968.metering.metering import *
from zepben.evolve.model.cim.iec61968.common.organisation import *
from zepben.evolve.model.cim.iec61968.common.document import *
from zepben.evolve.model.cim.iec61968.common.organisation_role import *
from zepben.evolve.model.cim.iec61968.common.location import *
from zepben.evolve.model.cim.iec61970.base.meas.control import *
from zepben.evolve.model.cim.iec61970.base.meas.measurement import *
from zepben.evolve.model.cim.iec61970.base.meas.value import *
from zepben.evolve.model.cim.iec61970.base.meas.iopoint import *
from zepben.evolve.model.cim.iec61970.base.diagramlayout.diagram_layout import *
from zepben.evolve.model.cim.iec61970.base.diagramlayout.orientation_kind import *
from zepben.evolve.model.cim.iec61970.base.diagramlayout.diagram_style import *
from zepben.evolve.model.cim.iec61970.base.diagramlayout.diagram_object_style import *
from zepben.evolve.model.cim.iec61970.base.scada.remote_point import *
from zepben.evolve.model.cim.iec61970.base.scada.remote_source import *
from zepben.evolve.model.cim.iec61970.base.scada.remote_control import *
from zepben.evolve.model.cim.iec61970.base.domain.unit_symbol import *
from zepben.evolve.model.cim.iec61970.base.auxiliaryequipment.auxiliary_equipment import *
from zepben.evolve.model.cim.iec61970.base.wires.line import *
from zepben.evolve.model.cim.iec61970.base.wires.energy_consumer import *
from zepben.evolve.model.cim.iec61970.base.wires.aclinesegment import *
from zepben.evolve.model.cim.iec61970.base.wires.per_length import *
from zepben.evolve.model.cim.iec61970.base.wires.vector_group import *
from zepben.evolve.model.cim.iec61970.base.wires.winding_connection import *
from zepben.evolve.model.cim.iec61970.base.wires.shunt_compensator import *
from zepben.evolve.model.cim.iec61970.base.wires.power_transformer import *
from zepben.evolve.model.cim.iec61970.base.wires.energy_source_phase import *
from zepben.evolve.model.cim.iec61970.base.wires.phase_shunt_connection_kind import *
from zepben.evolve.model.cim.iec61970.base.wires.connectors import *
from zepben.evolve.model.cim.iec61970.base.wires.switch import *
from zepben.evolve.model.cim.iec61970.base.wires.energy_source import *
from zepben.evolve.model.cim.iec61970.base.wires.single_phase_kind import *
from zepben.evolve.model.cim.iec61970.base.wires.energy_connection import *
from zepben.evolve.model.cim.iec61970.base.core.substation import *
from zepben.evolve.model.cim.iec61970.base.core.terminal import *
from zepben.evolve.model.cim.iec61970.base.core.equipment import *
from zepben.evolve.model.cim.iec61970.base.core.conducting_equipment import *
from zepben.evolve.model.cim.iec61970.base.core.identified_object import *
from zepben.evolve.model.cim.iec61970.base.core.base_voltage import *
from zepben.evolve.model.cim.iec61970.base.core.power_system_resource import *
from zepben.evolve.model.cim.iec61970.base.core.connectivity_node_container import *
from zepben.evolve.model.cim.iec61970.base.core.regions import *
from zepben.evolve.model.cim.iec61970.base.core.phase_code import *
from zepben.evolve.model.cim.iec61970.base.core.equipment_container import *
from zepben.evolve.model.cim.iec61970.base.core.connectivity_node import *
from zepben.evolve.model.cim.iec61970.infiec61970.feeder.circuit import *
from zepben.evolve.model.cim.iec61970.infiec61970.feeder.loop import *
from zepben.evolve.model.phasedirection import *
from zepben.evolve.model.phases import *


#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Optional

from zepben.evolve.model.cim.iec61970.base.meas.measurement import Measurement
from zepben.evolve.model.cim.iec61970.base.scada.remote_point import RemotePoint

__all__ = ["RemoteSource"]


class RemoteSource(RemotePoint):
    """
    Remote sources are state variables that are telemetered or calculated within the remote unit.
    """

    measurement: Optional[Measurement] = None
    """The `meas.measurement.Measurement` for the `RemoteSource` point."""
